/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

// On windows the cache is not written until firefox session ends with default config.
// To make cache files always be updated, open:
// about:config
// Look for these parameter and change as suggested
// browser.cache.memory.enable   false
// browser.cache.disk.enable     true
// browser.cache.offline.enable  false
//
// _CACHE_001_ will be updated in 6-7 seconds on my system and could now be monitored (... maybe)
// Better solution is to make a firefox addon to flush cache or to send it via dbus

//    tested and works with
//    http://www.youtube.com/  (treated as in native mediadownloader mode)
//    http://www.metacafe.com/
//    http://www.myspace.com/
//    http://www.stupidvideos.com/
//    http://www.museum.tv/
//    http://www.videoart.net/
//    http://espn.go.com/
//    http://it.sevenload.com/


//    to view cache into FireFox type:
//    about:cache?device=disk
//    The file structure of each _CACHE_00n file is as follows:

//     ______________________________
//    |                              |
//    |       4096 byte free         |
//    |        block bit-map         |
//    |------------------------------|
//    |                              |
//    |            Blocks            |
//    |             ...              |
//    |                              |
//    |                              |
//    |______________________________|
//    The size of each block varies depending on the file.
//    _CACHE_001_          256 byte blocks
//    _CACHE_002_         1024 byte blocks
//    _CACHE_003_         4096 byte blocks


//    Data in the cache is always encoded in a group of 1-4 blocks and always
//    starts exactly on a block boundary (an exact multiple of 256, 1024, or 4096 bytes).
//    For cache data that corresponds to metadata, it is encoded in this very specific data format.


//          _________________
//    0-3  |       magic      |    unsigned int   (first 4 bytes of _CACHE_MAP_)   (Big Endian)
//         |------------------|
//    4-7  |    location      |    unsigned int  (Big Endian)
//         |------------------|
//    8-11 |    fetchcount    |    unsigned int  (Big Endian)
//         |------------------|
//    12-15|    fetchtime     |    unsigned int  (Big Endian)
//         |------------------|
//    16-19|    modifytime    |    unsigned int  (Big Endian)
//         |------------------|
//    20-23|    expiretime    |    unsigned int  (Big Endian)
//         |------------------|
//    24-27|    datasize      |    unsigned int  (Big Endian)
//         |------------------|
//    28-31|    requestsize   |    unsigned int  (Big Endian)
//         |------------------|
//    32-35|    infosize      |    unsigned int  (Big Endian)
//         |------------------|
//    36-  |  request string  |    char[].  Length is determined by
//         |                  |    the value of the requestsize field
//         |       ...        |    above.  Terminated by a \x00 byte.
//         |                  |
//         |                  |
//         |------------------|
//    n-   |   info string    |    char[].  Length is determined by
//         |                  |    the value of the infosize field
//         |       ...        |    above.
//         |                  |
//         |                  |
//         |                  |
//         |__________________|



#include <QDebug>
#include <QDateTime>
#include <QFileInfo>
#include <QSettings>
#include <QFileDialog>
#include <QDesktopServices>
#include <QMessageBox>
#include <QProcess>
#include <QClipboard>

#include "firefoxmonitor.h"
#include "settings.h"
#include "youtube.h"

#ifdef Q_WS_X11
    //#include <X11/Xlib.h>
    #include <X11/keysym.h>
    //#include <X11/Xresource.h>
    //#include <X11/Xutil.h>
    #include <X11/extensions/XTest.h> // with libxtst-dev
#endif

#if defined(Q_WS_WIN) // || defined(Q_WS_PM)
    #include <windows.h>
#endif

#define TIMER_DEBUG 0
#define DEBUG_MIME_TYPE 0

firefoxMonitor::firefoxMonitor(ImageViewer2d *viewer, bool *canMonitor, QObject *parent) :
    QObject(parent)
{
    settings conf;

    if (conf.conf->value("firefox/path","").toString()=="") {
        if (getFirefoxCacheDir())
            conf.store_settings("firefox/path", firefoxCacheDir.absolutePath());
        else {
            QMessageBox::critical(0,QObject::tr("error"),
                                  QObject::tr("Cannot find firefox cache directory.\nYou can edit manually the mediadownloader.conf file and add an entry like:\n[firefox]\npath=PATH_TO_FIREFOX_CACHE") );
            *canMonitor=false;
            return;
        }
    }
    else {
        firefoxCacheDir.setPath( conf.conf->value("firefox/path","").toString() );
        // remove terminating "/Cache" directory
        QString ff = firefoxCacheDir.absolutePath();
#ifdef Q_WS_X11
        ff.chop( 6 );
        firefoxDir.setPath( ff );
#endif
#if defined(Q_WS_WIN) || defined(Q_WS_PM)
        firefoxDir.setPath(QString(getenv("APPDATA")) + "/Mozilla/Firefox");
#endif
    }
    *canMonitor=true;


    now   = QDateTime::currentDateTime().toTime_t();
    scene = viewer;

    // Set paths for database
#ifdef DEBUGGING
    mediadownloaderDir.setPath(QString(QUOTEME(PWD_PRO)));
#else
    #ifdef Q_WS_X11
        mediadownloaderDir.setPath(QDir::homePath()+"/.mediadownloader");
    #endif

    #if defined(Q_WS_WIN) || defined(Q_WS_PM)
        mediadownloaderDir.setPath(QDir::currentPath());
    #endif
#endif

    // Read first 4 bytes from _CACHE_MAP_ to get magic field
    QFile f(firefoxCacheDir.toNativeSeparators( firefoxCacheDir.absolutePath() + "/_CACHE_MAP_") );
    if (!f.open(QIODevice::ReadOnly))
    {
        *canMonitor=false;
        QMessageBox::critical(0,QObject::tr("error"),
                              QObject::tr("firefox cache directory found, but there is not _CACHE_MAP_ file.\nTry to start or restart firefox.") );
        conf.store_settings("firefox/path", QString(""));
        return;
    }
    bool b;
    magic = f.read(4).toHex().toULong(&b, 16);
    f.close();

    connect(&timer, SIGNAL(timeout()), this, SLOT(onTimeElapsed()) );

    cacheWatcher.addPath(firefoxCacheDir.toNativeSeparators( firefoxCacheDir.absolutePath() + "/_CACHE_001_" ));
    connect(&cacheWatcher, SIGNAL(fileChanged(QString)), this, SLOT(onFirefoxCacheChanged(QString)) );
//    qDebug() << "CACHE_001 " << firefoxCacheDir.toNativeSeparators( firefoxCacheDir.absolutePath() + "/_CACHE_001_");
}



firefoxMonitor::~firefoxMonitor()
{

}

// Ref: http://kb.mozillazine.org/Profile_folder
// For now it works only for single firefox profile
bool firefoxMonitor::getFirefoxCacheDir()
{
    // looking for firefox path
#ifdef Q_WS_X11
    firefoxDir.setPath(QDir::homePath()+"/.mozilla/firefox");
    qDebug() << firefoxDir;
#endif

#if defined(Q_WS_WIN) || defined(Q_WS_PM)
//    Windows NT (NT4.x/2000/XP/Vista/7)   "%APPDATA%\Mozilla\"
//    c:/Users/deimos/Application Data/Mozilla/Firefox

    firefoxDir.setPath(QString(getenv("APPDATA")) + "/Mozilla/Firefox");
#endif
    if (!firefoxDir.exists())
    {
        // if not found ask to user where to look
        QMessageBox::warning(0,QObject::tr("warning"),
                             QObject::tr("Mediadownloader can't find firefox cache directory.\n"
                                         "Try to manually look for that.\nIf you don't know where it is, "
                                         "write in firefox address url this: about:cache?device=disk\n"
                                         "The needed directory is written beside \"Cache Directory:\"\n"
                                         "You can try to locate it yourself in the follows up dialog or"
                                         "open mediadownloader.conf and add an entry like this:\n[firefox]\npath=DIR_FOUND") );
        QFileDialog dialog;
        QString msg;
        QList<QUrl> sideUrls;
        sideUrls << QUrl::fromLocalFile(QDir::homePath()) << QUrl::fromLocalFile(QDesktopServices::storageLocation(QDesktopServices::DesktopLocation));
        dialog.setFileMode(QFileDialog::Directory);
        dialog.setViewMode(QFileDialog::List);
        dialog.setDirectory(QDir::homePath());
        msg=tr("Select firefox files directory\nUsually in");
#ifdef Q_WS_X11
        msg += " HOME/.mozilla";
#else
        msg += " %APPDATA%\\Mozilla\\";
        sideUrls << QUrl::fromLocalFile(QDesktopServices::storageLocation(QDesktopServices::ApplicationsLocation));
#endif
        dialog.setFilter(QDir::Hidden | QDir::AllDirs );
        dialog.setSidebarUrls(sideUrls);
        dialog.setLabelText(QFileDialog::FileName, msg);

        QStringList s;
        if (dialog.exec()) {
            s = dialog.selectedFiles();
            if (!firefoxDir.exists()) {
                if (s.size()!=0)
                    firefoxDir.setPath(s.at(0));
            }
            else {
                // If we are here is because directories (cache dir and files dir) in windows are different
                qDebug() << firefoxCacheDir;
                firefoxCacheDir.setPath(s.at(0));
            }
        }
    }

    // if found look for profiles.ini to get into it where "Cache" dir is
    if (firefoxDir.exists()) {
        QString cachePath;
        QFile f( QDir::toNativeSeparators( firefoxDir.absolutePath()+"/profiles.ini") );
        if (!f.exists()) {
            QMessageBox::critical(0,QObject::tr("error"),QObject::tr("Cannot find profiles.ini") );
            return false;
        }
        else {
            QSettings settings(f.fileName(), QSettings::IniFormat);
            cachePath = settings.value("Profile0/Path","").toString();

            firefoxCacheDir = firefoxDir.toNativeSeparators( firefoxDir.absolutePath() ) +
                    "/" + cachePath + "/Cache";

            if (firefoxCacheDir.exists()) qDebug() << "Found firefox cache directory at: " << firefoxCacheDir.absolutePath();
        }
    }
    else {
        QMessageBox::critical(0,QObject::tr("error"),QObject::tr("Cannot find firefox directory") );
        return false;
    }
    return true;
}

void firefoxMonitor::onFirefoxCacheChanged(QString file)
{
    Q_UNUSED(file)
    if (TIMER_DEBUG) qDebug() << "TIMER STARTED";
    timer.start(3000);
}

void firefoxMonitor::onTimeElapsed()
{
    if (cacheWatcher.files().size()==0)
    {
        if (TIMER_DEBUG) qDebug() << "ADDED monitor";
        cacheWatcher.addPath( firefoxCacheDir.toNativeSeparators( firefoxCacheDir.absolutePath() + "/_CACHE_001_") );
    }
    if (TIMER_DEBUG) qDebug() << "TIMER STOPPED";
    if (!readFirefoxUrl())
        readCache();
    timer.stop();
}

void firefoxMonitor::readCache()
{
    QFile f(firefoxCacheDir.toNativeSeparators( firefoxCacheDir.absolutePath() + "/_CACHE_001_") );
    cache_data record;

    if (!f.open(QIODevice::ReadOnly))
        return;
    bool b=f.seek(4096);

    while (!f.atEnd()) {
        record.magic       = f.read(4).toHex().toULong(&b, 16);
        record.location    = f.read(4).toHex().toULong(&b, 16);
        record.fetchcount  = f.read(4).toHex().toULong(&b, 16);
        record.fetchtime   = f.read(4).toHex().toULong(&b, 16);
        record.modifytime  = f.read(4).toHex().toULong(&b, 16);
        record.expiretime  = f.read(4).toHex().toULong(&b, 16);
        record.datasize    = f.read(4).toHex().toULong(&b, 16);
        record.requestsize = f.read(4).toHex().toULong(&b, 16);
        record.infosize    = f.read(4).toHex().toULong(&b, 16);

        if (record.magic==magic)
        {
            record.requestString  = f.read(record.requestsize);
            record.infoString     = f.read(record.infosize).replace((char)0x00, (char)0x20).replace("\r\n","\n");
            addRecord(record);
        }
        // Calculate 256 byte bounduary
        f.seek(f.pos()+ (256-(f.pos()%256)) );
    }
    f.close();
}

void firefoxMonitor::addRecord(cache_data record)
{

    // These octect-stream work on some sites and not on others
    // BTW are used also for banners so leave them for now
//  "Content-Type: application/octet-stream"
//  "Content-Type: binary/octet-stream"

    // Check if time is after we started monitoring
    if (record.fetchtime<=now) return;

    // Check if there is a recent video content
    QString contentType;        // like "video/x-flv"
    QString contentTypeKind;    // like "x-flv"

    if (DEBUG_MIME_TYPE) {
        QRegExp rx("Content-Type:.*\\n");
        rx.setMinimal(true);
        if (rx.indexIn(record.infoString, 0) != -1)
            qDebug() << rx.cap(0);
    }

//    QRegExp rx("Content-Type: (video/(.+))\\n");
    QRegExp rx("Content-Type: ((video|application)/(.*))\\n");
    rx.setMinimal(true);

    if (rx.indexIn(record.infoString, 0) != -1 )
    {
        contentType     = rx.cap(1);
        contentTypeKind = rx.cap(3);
    }
    else return;

    if (contentType.contains("application") && !contentType.contains("x-shockwave-flash"))
        return;

    emit outMessage(tr("Video found: ") + contentType , 1);

    if (contentType=="application/x-shockwave-flash")
    {
        emit outMessage(tr("but swf is not supported") ,2);
        return;
    }
    else
        emit outMessage("" ,2);

    if (!( contentTypeKind.contains("flv")     ||
           contentTypeKind.contains("mp4")     ||
           contentTypeKind.contains("msvideo") ||
           contentTypeKind.contains("mpeg")    ||
           contentTypeKind.contains("quicktime")  )
            )
    {
        emit outMessage(tr("but maybe cannot play this format") ,2);
    }

    record.requestString=QByteArray::fromPercentEncoding(record.requestString).remove(0,5);

    // If it is an adv link retrun
    if (record.requestString.contains("doubleclick"))
        return;

    // If it is a youtube link treat it like a native Mediadownloader type
    if (record.requestString.contains("youtube")) {
        readFirefoxUrl();
        return;
    }


    // Checks if already present
    imageContainer *ic;
    foreach (ic, scene->itemsList) {
        if (ic->image.image_url == record.requestString) return;
    }

    if (TIMER_DEBUG) qDebug() << record.requestString;
    if (TIMER_DEBUG) qDebug() << record.infoString;
    if (TIMER_DEBUG) qDebug() << "============================================================";

    img.kind = CONTAINER_KIND_VIDEO_OTHER;
    img.image_tburl = "";
    img.image_url = record.requestString;
    img.ID=scene->n_widget;

    // When saving,  take care of img.description_row2 for file extension
    // in save::download_one
    img.description_row1=QUrl(record.requestString).encodedHost();
    img.description_row2=contentType;


    rx.setPattern("Content-Length: ([0-9]{1,})");
    if (rx.indexIn(record.infoString, 0) != -1)
        img.description_row3= QString::number(rx.cap(1).toULong()/1024) +" Kb";

    scene->add_imageContainer("", img);
    scene->addItemList_to_scene();
    now   = QDateTime::currentDateTime().toTime_t();
}

bool firefoxMonitor::readFirefoxUrl()
{
    // Fire up CTRL+L and CTRL+C to fireFox to copy url into clipboard
#ifdef Q_WS_X11
    // Obtain the X11 display.
    Display *display = XOpenDisplay(0);
    if(display == NULL)
        return false;

    // Get window id:
    //    xprop -root | awk '/_NET_ACTIVE_WINDOW\(WINDOW\)/{print $NF}'
    // Get pid:
    //    xprop -id $(xprop -root | awk '/_NET_ACTIVE_WINDOW\(WINDOW\)/{print $NF}') | awk '/_NET_WM_PID\(CARDINAL\)/{print $NF}'

    // Get firefox pid
    QProcess p;
    QString cmd( "/bin/sh" );
    QString s;
    QStringList args;
    args << "-c" << "ps -A| grep firefox";
    p.start(cmd, args);
    p.waitForReadyRead();
    s = p.readAll();
    p.waitForFinished();
    QRegExp rx("(\\d+.) ");
    rx.setMinimal(true);
    if (rx.indexIn(s, 0) == -1) return false;

    QString firefoxPID = rx.cap(1);
    // Get top level window pid
    args.clear();
    args << "-c" << "xprop -id $(xprop -root | awk '/_NET_ACTIVE_WINDOW\\(WINDOW\\)/{print $NF}') | awk '/_NET_WM_PID\\(CARDINAL\\)/{print $NF}'";
    p.start(cmd, args);
    p.waitForReadyRead();
    s = p.readAll().replace("\n","");
    p.waitForFinished();
    if (s != firefoxPID) return false;

    // Find the window which has the current keyboard focus.
//    Window winFocus;
//    int    revert;
//    XGetInputFocus(display, &winFocus, &revert);
//    XFlush(display);
//    qDebug() << " WINDOW winFocus " << winFocus;


    // If pid of the focused window is the same of firefox we can procede to send keys
    KeyCode keyCTRL = XKeysymToKeycode(display, XK_Control_L);
    KeyCode keyL = XKeysymToKeycode(display, XK_l);
    KeyCode keyC = XKeysymToKeycode(display, XK_c);

    XTestFakeKeyEvent(display, keyCTRL, True, 0);
    XTestFakeKeyEvent(display, keyL, True, 0);
    XTestFakeKeyEvent(display, keyL, False, 0);
    XTestFakeKeyEvent(display, keyCTRL, False, 0);

    XTestFakeKeyEvent(display, keyCTRL, True, 0);
    XTestFakeKeyEvent(display, keyC, True, 0);
    XTestFakeKeyEvent(display, keyC, False, 0);
    XTestFakeKeyEvent(display, keyCTRL, False, 0);

    // Done.
    XCloseDisplay(display);
#endif

#if defined(Q_WS_WIN) || defined(Q_WS_PM)
//    wchar_t KeyL='l';
//    INPUT input[2];
//    input[0].ki.wVk = VK_CONTROL;
//    input[0].ki.dwFlags = 0;
//    input[0].ki.wVk = 0;
//    input[0].ki.wScan = KeyL;
//    input[0].ki.dwFlags = KEYEVENTF_UNICODE;

//    input[1].ki.wVk = VK_CONTROL;
//    input[1].ki.dwFlags = KEYEVENTF_KEYUP;
//    input[1].ki.wVk = 0;
//    input[1].ki.wScan = KeyL;
//    input[1].ki.dwFlags = KEYEVENTF_UNICODE | KEYEVENTF_KEYUP;

//    SendInput(6, input, sizeof(INPUT));

#endif

    // Copy clipboard content
    // Clear clipboard
    QClipboard *clipboard = QApplication::clipboard();
    QString url = clipboard->text();
    clipboard->clear();
    if (!url.contains(("youtube"))) return false;
    addYouTubeByUrl(url);
    return true;
}


void firefoxMonitor::addYouTubeByUrl(QString url)
{
    QRegExp rx("v=(.*)[&]{0,}$");
    rx.setMinimal(true);
    if (rx.indexIn(url, 0) != -1) {
        QList<YouTubeStruct> tubes;
        QString youtubeID = rx.cap(1);
        qDebug() << rx.cap(1);
        // Check if the ID is already processed
        if (scene->n_widget!=0) {
            for (int i=0; i < scene->itemsList.size(); i++) {
                if (scene->itemsList.at(i)->image.youTubeInfo.video_id.contains(youtubeID)) {
                    if (TIMER_DEBUG) qDebug() << "VIDEOID already there ";
                    return;
                }
            }
        }
        // If it is a new video do a standard search
        tubes.clear();
        tubes=youTube::getList(youtubeID);
        if (tubes.size()>0)
        {
            QList<imageStruct> imgs;
            img.kind = CONTAINER_KIND_VIDEO;
            img.youTubeInfo=tubes.at(0);
            img.youTubeInfo.video_id=youtubeID;
            img.description_row1=tubes.at(0).title;
            img.description_row2=tubes.at(0).content;
            img.description_row3=tubes.at(0).duration+"\"";
            img.image_tburl=tubes.at(0).p[3].url;  //4th thumbnail is that one with better resolution
            img.image_url=tubes.at(0).id;
            img.image_source=tubes.at(0).mediaUrl;
            imgs << img;
            scene->add_all_images(imgs);
            qDebug() << "YouTube video added: " << youtubeID;
            now   = QDateTime::currentDateTime().toTime_t();
        }
    }
}


// This is a try but didn't work well... I live it commented, maybe usefull for some other things

/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
// check firefox places.sqlite database for youtube urls
// Firefox take exclusive rights for that database, so must be closed
// before this
//bool firefoxMonitor::initDbase()
//{
//    if (!QSqlDatabase::isDriverAvailable("QSQLITE"))
//    {
//        qDebug() << "SQLite 3.x driver unavailabe. Check Qt build!";
//        return false;
//    }


//    // copy the database to make it unlocked
////    qDebug() << "COPYING  : " << firefoxDir.absolutePath()+"/places.sqlite";
////    qDebug() << "TO       : " << QDir::toNativeSeparators(mediadownloaderDir.absolutePath()) + "/places.sqlite";
//    bool b=QFile::remove( QDir::toNativeSeparators(mediadownloaderDir.absolutePath()) + "/places.sqlite" );
//    if (! QFile::copy( firefoxDir.absolutePath()+"/places.sqlite" ,
//                       QDir::toNativeSeparators(mediadownloaderDir.absolutePath()) + "/places.sqlite" ) )
//    {
//        qDebug() << "CONNOT copy sqlite";
//        return false;
//    }



//    db = QSqlDatabase::addDatabase("QSQLITE","moz_places");
////    db.setConnectOptions("QSQLITE_OPEN_READONLY=1");
//    db.setDatabaseName(QDir::toNativeSeparators(mediadownloaderDir.absolutePath()) + "/places.sqlite");
//    bool ok=db.open();
//    if (!ok) {
//        qDebug() << "CANNOT OPEN dbase";
//        return false;
//    }

//    return true;
//}

//bool firefoxMonitor::checkForNewYoutubeVideo()
//{
//    bool ret=true;
//    QString url;

//    if (!initDbase()) {qDebug() << "!INITDBASE"; ret = false;}
//    if (ret) {
//        QSqlQuery *query = new QSqlQuery(db);
//        // Get last field and check for youtube video id
//        // http://www.youtube.com/watch?v=HaEzY2RCGG0&feature=related
//        //    query->exec("SELECT url, last_visit_date FROM moz_places WHERE url LIKE '%youtube%' AND last_visit_date>="+QString::number(now)+"*1000000 ORDER BY id ASC");
//        query->exec("SELECT url, last_visit_date FROM moz_places WHERE url LIKE '%youtube%' ORDER BY id DESC");

//        qDebug() <<  "Query        : " << query->lastQuery();
//        qDebug() <<  "Query error  : " << query->lastError().text();
//        qDebug() <<  "Query results: " << query->size();
//        QString youtubeID;
//        QList<YouTubeStruct> tubes;

//        if (query->first()) {
//            url = query->value(0).toString();
//            qDebug() << "URL: " << url;
//            QRegExp rx("v=(.+)&");
//            rx.setMinimal(true);
//            if (rx.indexIn(url, 0) != -1) {
//                youtubeID = rx.cap(1);
//                qDebug() << "YOUTUBEID: " << youtubeID;
//                // Check if the ID is already processed
//                if (scene->n_widget!=0) {
//                    for (int i; i <= scene->itemsList.size(); i++) {
//                        if (scene->itemsList.at(i)->image.youTubeInfo.video_id.contains(youtubeID)) { qDebug() << "VIDEOID already there ";ret = false; break; }
//                    }
//                }
//                if (ret) {
//                    // If it is a new video do a standard search
//                    tubes.clear();
//                    qDebug() << "GETLIST " << youtubeID;
//                    tubes=youTube::getList(youtubeID);
//                    if (tubes.size()>0)
//                    {
//                        QList<imageStruct> imgs;
//                        img.kind = CONTAINER_KIND_VIDEO;
//                        img.youTubeInfo=tubes.at(0);
//                        img.youTubeInfo.video_id=youtubeID;
//                        img.description_row1=tubes.at(0).title;
//                        img.description_row2=tubes.at(0).content;
//                        img.description_row3=tubes.at(0).duration+"\"";
//                        img.image_tburl=tubes.at(0).p[3].url;  //4th thumbnail is that one with better resolution
//                        img.image_url=tubes.at(0).id;
//                        img.image_source=tubes.at(0).mediaUrl;
//                        imgs << img;
//                        scene->add_all_images(imgs);
//                        qDebug() << "YOUTUBE ADDED";
//                    }
//                }
//            }
//        }
//        delete query;
//    }
//    //    db.commit();
////    db.removeDatabase("moz_places");
//    {
//        QSqlDatabase database = QSqlDatabase::database("moz_places");
//        database.close();
//    }
//    QSqlDatabase::removeDatabase("moz_places");
//    QFile::remove( QDir::toNativeSeparators(mediadownloaderDir.absolutePath()) + "/places.sqlite" );
//    if (url.isEmpty()) ret = false;
//    qDebug() << "RET = " << ret << ": " << url;
//    return ret;
//}

