/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

#include <QDebug>
#include <QMutexLocker>
#include <QDir>
#include <QApplication>

#include "mbbuffer.h"

#define DEBUG_IODEVICE 0   // debug QIODevice reimplemented functions
#define DEBUG_NAM_THREAD 0 // debug MBbufferThread
#define DEBUG_NAM 0        // debug getAsync and getSync

#define BUFFER_SIZE 200LL*1024

//    It's not possible to use QNetworkAccessManager in a thread on Maemo with Qt
//    4.7.
//    http://bugreports.qt.nokia.com/browse/QTBUG-15004
//    https://bugs.kde.org/show_bug.cgi?id=257376

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
// with async=true bufferEmptyed() and bufferReady() signals are emitted
MBbufferThread::MBbufferThread(bool asynchronous, QObject *parent) :
    QObject(parent),
    async(asynchronous)
{
//    manager.moveToThread(QApplication::instance()->thread());
    manager = new QNetworkAccessManager(this);
}

MBbufferThread::~MBbufferThread()
{
    if (DEBUG_NAM_THREAD) qDebug() << "DELETE MBbufferThread" << this;
    if (!reply.isNull())
        reply->abort();
    delete manager;
}

void MBbufferThread::init(QUrl url, MBbuffer *device, qint64 rangeStart, qint64 rangeEnd, qint64 startAlreadyDownloaded)
{
    currentUrl = url;
    buffer = device;
    startPos = rangeStart;
    currWritePos = rangeStart;
    endPos = rangeEnd;
    alreadyDownloadedPos = startAlreadyDownloaded;
    bufferReadyEmitted = false;
    resetBuffer();
}

void MBbufferThread::abortNAM()
{
    if (DEBUG_NAM_THREAD) qDebug() << "ABORT MBbufferThread";
    if (!reply.isNull())
        reply->abort();
    emit getFinished();
}

bool MBbufferThread::NAMisRunning()
{
    if (reply.isNull()) return false;
    return reply->isRunning();
}

qint64 MBbufferThread::lowerWrittenPos()
{
    return alreadyDownloadedPos;
}

qint64 MBbufferThread::currentWritingPos()
{
    return currWritePos;
}

qint64 MBbufferThread::highRangeValue()
{
    return endPos;
}

qint64 MBbufferThread::lowRangeValue()
{
    return alreadyDownloadedPos;
}

void MBbufferThread::resetBuffer()
{
    bufferReadyEmitted=false;
    if (async) emit bufferEmptyed();
}

void MBbufferThread::setNAMcookies(MbCookieJar *cookieJar)
{
    cookies = cookieJar;
}

void MBbufferThread::run()
{
    QNetworkRequest request(currentUrl);

    request.setRawHeader("User-Agent", "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b11) Gecko/20100101 Firefox/4.0b11");
    request.setRawHeader("Accept-Charset", "ISO-8859-1,utf-8;q=0.7,*;q=0.7");
    request.setRawHeader("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
    request.setRawHeader("Accept-Encoding", "gzip, deflate");
    request.setRawHeader("Accept-Language", "en-us,en;q=0.5");
    request.setRawHeader("Connection", "Keep-Alive");
    if (!(startPos==0 && endPos==0) && (endPos>startPos+1)) {
        QByteArray arg;
        arg = "bytes=";
        arg += QString::number(startPos)+"-";
        arg += QString::number(endPos);
        request.setRawHeader("Range", arg );
    }

    tmpCookieJar = new MbCookieJar;
    tmpCookieJar->copyFrom(cookies);
//    qDebug() << "1-MBbufferThread" << this->thread() << "manager" << manager->thread();
#ifdef MOBILE_APP  // try to keep manager to main thread on maemo for the bug http://bugreports.qt.nokia.com/browse/QTBUG-15004
    manager->moveToThread(this->thread());
#endif
//    qDebug() << "2-MBbufferThread" << this->thread() << "manager" << manager->thread();
    if (tmpCookieJar!=0) manager->setCookieJar(tmpCookieJar);
    reply = manager->get(request);
    qRegisterMetaType<QNetworkReply::NetworkError>("QNetworkReply::NetworkError");

    QObject::connect(reply.data(), SIGNAL(finished()),
            this, SLOT(downloadFinished()) );
    QObject::connect(reply.data(), SIGNAL(downloadProgress(qint64,qint64)),
            this, SLOT(downloadProgress(qint64,qint64)) );
    QObject::connect(reply.data(), SIGNAL(error(QNetworkReply::NetworkError)),
            this, SLOT(downloadError(QNetworkReply::NetworkError)) );
    QObject::connect(reply.data(), SIGNAL(sslErrors(QList<QSslError>)),
            this, SLOT(abortNAM()) );
    QObject::connect(reply.data(), SIGNAL(destroyed()),
            this, SLOT(abortNAM()) );
    bufferReadyEmitted=false;
//    if (async) emit bufferEmptyed();
}

// TODO: add metadataChanged
void MBbufferThread::downloadProgress(qint64 cur,qint64 tot)
{
    Q_UNUSED(cur)
    Q_UNUSED(tot)
    if (currWritePos > buffer->size()  )
        this->abortNAM();

    if ( !reply.isNull()) {
        QByteArray content;
        content = reply->readAll();
        if (content.size()>0) {
            buffer->seekWriteFile(currWritePos);
            qint64 n=buffer->write(content, content.size());
            currWritePos += n;
            if (async) {
                if (!bufferReadyEmitted && currWritePos-buffer->readPosValue() >= BUFFER_SIZE) {
                    if (DEBUG_NAM_THREAD) qDebug() << "BUFFERED " << currWritePos-buffer->readPosValue() << "\n";
                    bufferReadyEmitted=true;
                    emit bufferReady();
                }
                if (bufferReadyEmitted && currWritePos-buffer->readPosValue() < BUFFER_SIZE) {
                    if (DEBUG_NAM_THREAD) qDebug() << "BUFFER EMPTYED";
                    bufferReadyEmitted=false;
                    emit bufferEmptyed();
                }
            }
        }
    }
}

void MBbufferThread::downloadFinished()
{
    if (DEBUG_NAM_THREAD) qDebug() << "downloadFinished [" << startPos << currWritePos << endPos << "]" << this;
    emit getFinished();
}

void MBbufferThread::downloadError(QNetworkReply::NetworkError e)
{
    if (DEBUG_NAM_THREAD) qDebug() << "download error: " << e << reply->errorString()  << this;
    emit getError();
}
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
bool MBbuffer::singleton=false;
MBbuffer::MBbuffer(QUrl url, MbCookieJar *cookieJar, QObject *parent) :
    QIODevice(parent)
{
    singleton=true;
    setNAMcookies(cookieJar);
    init(url);
}

MBbuffer::~MBbuffer()
{
    QMutexLocker rlocker(&readLock);
    QMutexLocker wlocker(&writeLock);
    if (bufferTimer.isActive()) bufferTimer.stop();
    if ( getAsyncThread!=0 ) delete getAsyncThread;
    emit aboutToClose();
    cacheFileWriteOnly.close();
    cacheFileWriteOnly.remove();
    singleton=false;
}

void MBbuffer::init(QUrl url)
{
    getAsyncThread=0;
    readPos = 0;
    writePos = 0;
    contentLength = 0;
    downloadSize = 0;
    contentBytesWritten = 0;
    if (cacheFileWriteOnly.isOpen()) cacheFileWriteOnly.close();
    if (cacheFileReadOnly.isOpen()) cacheFileReadOnly.close();

    if (cacheFileWriteOnly.exists()) cacheFileWriteOnly.remove();
    if (cacheFileReadOnly.exists()) cacheFileReadOnly.remove();

    cacheFileWriteOnly.setFileName(QDir::temp().absoluteFilePath("md_tmp"));
    cacheFileReadOnly.setFileName(QDir::temp().absoluteFilePath("md_tmp"));

    open(QIODevice::ReadWrite);
    setUrl(url);

    getContentLenght();

    bufferTimer.setSingleShot(true);
    buffering=false;
}

////////////////////////////////////////////
// Download  ///////////////////////////////
void MBbuffer::setUrl(QUrl url)
{
    currentUrl = url;
}

void MBbuffer::setNAMcookies(MbCookieJar *cookieJar)
{
    cookies = cookieJar;
}

void MBbuffer::seekWriteFile(qint64 pos)
{
    cacheFileWriteOnly.seek(pos);
    writePos=pos;
}

qint64 MBbuffer::getContentLenght()
{
    QEventLoop q;
    QNetworkRequest request(currentUrl);

    request.setRawHeader("User-Agent", "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b11) Gecko/20100101 Firefox/4.0b11");
    request.setRawHeader("Accept-Charset", "ISO-8859-1,utf-8;q=0.7,*;q=0.7");
    request.setRawHeader("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
    request.setRawHeader("Accept-Encoding", "gzip, deflate");
    request.setRawHeader("Accept-Language", "en-us,en;q=0.5");
    request.setRawHeader("Connection", "Keep-Alive");

    QNetworkAccessManager *manager = new QNetworkAccessManager(this);
    tmpCookieJar = new MbCookieJar;
    tmpCookieJar->copyFrom(cookies);
#ifdef MOBILE_APP  // on Maemo QNetworkAccessManager must not to be in another thread
    manager->moveToThread(this->thread());
#endif
    if (tmpCookieJar!=0) manager->setCookieJar(tmpCookieJar);
    reply = manager->head(request);
    connect(reply, SIGNAL(finished()), &q, SLOT(quit()) );
    connect(reply, SIGNAL(error(QNetworkReply::NetworkError)), &q, SLOT(quit()) );

    connect(reply, SIGNAL(destroyed()), &q, SLOT(quit()) );
    connect(reply, SIGNAL(readChannelFinished()), &q, SLOT(quit()) );
    connect(reply, SIGNAL(metaDataChanged()), &q, SLOT(quit()) );

    q.exec();

    QByteArray header = reply->rawHeader("Content-Length");
    contentLength = header.toLongLong();
    fillWithZeros(contentLength);
    downloadSize = contentLength;
    delete manager;
    emit totalSizeAcquired(contentLength);
    return contentLength;
}

void MBbuffer::getSync(qint64 start, qint64 end)
{
    if (end <= start) return;
    if (!cacheFileReadOnly.isOpen()) return;

    QMutexLocker getlocker(&getLock);
    QEventLoop q;
    MBbufferThread *getThread = new MBbufferThread(false);
    getThread->init(currentUrl, this, start, end, start);
    tmpCookieJar = new MbCookieJar;
    tmpCookieJar->copyFrom(cookies);
    getThread->setNAMcookies(tmpCookieJar);
    if (DEBUG_NAM) qDebug() << "MBbuffer::getSync" << start << end;
    connect(getThread, SIGNAL(getFinished()) , &q, SLOT(quit()) );
    connect(getThread, SIGNAL(getError()) , &q, SLOT(quit()) );
    connect(getThread, SIGNAL(destroyed()) , &q, SLOT(quit()) );
    QTimer::singleShot(1000, &q, SLOT(quit()) );
    getThread->run();
    if (getThread->NAMisRunning()) q.exec();
    cacheFileWriteOnly.flush();
//    getThread->exit();
    delete getThread;
//    getThread->deleteLater();
}

void MBbuffer::getAsync(qint64 start, qint64 end)
{
    if (!cacheFileReadOnly.isOpen()) return;
    QMutexLocker getlocker(&getLock);
    if (end <= start) return;

    if (buffering) return;

    // Check if already downloading
    if ( getAsyncThread!=0 &&
         ( start >= getAsyncThread->lowerWrittenPos() &&
           start < getAsyncThread->currentWritingPos() ) )
    {
        if (DEBUG_NAM) qDebug() << "Already downloading"  << "readpos" << readPos << "[" << getAsyncThread->lowerWrittenPos() << "---" << getAsyncThread->lowRangeValue() << getAsyncThread->currentWritingPos() << getAsyncThread->highRangeValue() << "]" << getAsyncThread;
        return;
    }

    // if already downloading, stop it (only 1 instance allowed)
    if (getAsyncThread != 0 && getAsyncThread->NAMisRunning()) {
        if (DEBUG_NAM) qDebug() << "STOPPING ASYNC" << "readpos" << readPos << "[" << getAsyncThread->lowerWrittenPos() << "---" << getAsyncThread->lowRangeValue() << getAsyncThread->currentWritingPos() << getAsyncThread->highRangeValue() << "]" << getAsyncThread;
        QTimer timer;
        QEventLoop q;
        QObject::connect(getAsyncThread, SIGNAL(getFinished()) , &q, SLOT(quit()) );
        QObject::connect(getAsyncThread, SIGNAL(bufferReady()) , &q, SLOT(quit()) );
//        QObject::connect(getAsyncThread, SIGNAL(terminated())  , &q, SLOT(quit()) );
        QObject::connect(getAsyncThread, SIGNAL(getError())    , &q, SLOT(quit()) );
        timer.singleShot( 10, getAsyncThread, SLOT(abortNAM()) );

        q.exec();
        timer.stop();
//        getAsyncThread->exit();
        delete getAsyncThread;
        getAsyncThread = 0;
    }
    getAsyncThread = new MBbufferThread(true);
    QEventLoop q;
    QObject::connect(getAsyncThread, SIGNAL(getFinished()) , &q, SLOT(quit()) );
    QObject::connect(getAsyncThread, SIGNAL(bufferReady()) , &q, SLOT(quit()) );
//    QObject::connect(getAsyncThread, SIGNAL(terminated())  , &q, SLOT(quit()) );
    QObject::connect(getAsyncThread, SIGNAL(getError())    , &q, SLOT(quit()) );

    QObject::connect(getAsyncThread, SIGNAL(bufferEmptyed()), this, SLOT(onBufferEmptyed()), Qt::DirectConnection);

    tmpCookieJar = new MbCookieJar;
    tmpCookieJar->copyFrom(cookies);
    getAsyncThread->setNAMcookies(tmpCookieJar);
    getAsyncThread->init(currentUrl, this, start, end, start);
    if (DEBUG_NAM) qDebug() << "RUNNING ASYNC" << "readpos" << readPos << "[" << getAsyncThread->lowerWrittenPos() << "---" << getAsyncThread->lowRangeValue() << getAsyncThread->currentWritingPos() << getAsyncThread->highRangeValue() << "]" << getAsyncThread;

    buffering=true;
    getAsyncThread->run();
    QObject::connect(getAsyncThread, SIGNAL(bufferReady()), this, SLOT(onBufferReady()), Qt::DirectConnection );

    q.exec();
//    cacheFileWriteOnly.flush();

    buffering=false;

    if (DEBUG_NAM) qDebug() << "RUNNING ASYNC OK" << "readpos" << readPos << "[" << getAsyncThread->lowerWrittenPos() << "---" << getAsyncThread->lowRangeValue() << getAsyncThread->currentWritingPos() << getAsyncThread->highRangeValue() << "]" << getAsyncThread;
    if (DEBUG_NAM) qDebug() << "\n";

}

void MBbuffer::onBufferReady()
{
    // With bufferTimer prevent to emit bufferReady and bufferEmptyed many times, at
    // last 1  every 1500 ms
    QObject::disconnect(&bufferTimer, 0,0,0);
    if (bufferTimer.isActive()) {
        bufferTimer.stop();
    }
    QObject::connect(&bufferTimer, SIGNAL(timeout()), this, SIGNAL(bufferReady()), Qt::DirectConnection );
    bufferTimer.start(1500);
}

void MBbuffer::onBufferEmptyed()
{
    QObject::disconnect(&bufferTimer, 0,0,0);
    if (bufferTimer.isActive()) {
        bufferTimer.stop();
    }
    QObject::connect(&bufferTimer, SIGNAL(timeout()), this, SIGNAL(bufferEmptyed()), Qt::DirectConnection );
    bufferTimer.start(1500);
}

void MBbuffer::startDownload()
{
    if (!isOpen()) {
        if (!open(QIODevice::ReadWrite)) qDebug() << "Open buffer failed";
    }
    getAsync(0, contentLength);
}

void MBbuffer::startDownloadFromBytePos(qint64 fromPos)
{
    if ( getAsyncThread!=0 &&
         !( fromPos >= getAsyncThread->lowerWrittenPos() &&
           fromPos < getAsyncThread->currentWritingPos() ) )
    {
        getAsync(fromPos, contentLength);
    }
}

void MBbuffer::startDownloadFromPerc(int perc)
{
    qint64 fromPos;
    fromPos = (qint64)contentLength/100*(perc-1);
    if ( getAsyncThread!=0 &&
         !( fromPos >= getAsyncThread->lowerWrittenPos() &&
           fromPos < getAsyncThread->currentWritingPos() ) )
    {
        getAsync(fromPos, contentLength);
    }
}

bool MBbuffer::readPosInRange(qint64 pos)
{
    return getAsyncThread!=0 &&
           ( pos >= getAsyncThread->lowerWrittenPos() &&
             pos < getAsyncThread->currentWritingPos() );
}

bool MBbuffer::isBuffering()
{
    return buffering;
}

////////////////////////////////////////////
////////////////////////////////////////////
qint64 MBbuffer::fillWithZeros(qint64 size)
{
    qint64 ret;
    char *c;
    size = size;
    c = (char*)malloc(size);
    memset(c,0,size);
    ret = cacheFileWriteOnly.write(c, size);
    if (DEBUG_IODEVICE) qDebug() << "StreamBuffer::fillWithZeros" << size;
    free(c);
    contentLength = size;
    writePos = 0;
    readPos = 0;

    return ret;
}

qint64 MBbuffer::readPosValue()
{
    return readPos;
}

////////////////////////////////////////////
////////////////////////////////////////////
// Reimplemented QIODevice funtions ////////
////////////////////////////////////////////
bool MBbuffer::isSequential() const
{
    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::isSequential";
    return false;
}

bool MBbuffer::atEnd() const
{
    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::atEnd" << writePos << readPos;
    return (readPos == contentLength);
}
void MBbuffer::close()
{
    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::close";
    cacheFileWriteOnly.close();
    QIODevice::close();
}

bool MBbuffer::open( OpenMode mode )
{
    Q_UNUSED(mode)
    QMutexLocker rlocker(&readLock);
    QMutexLocker wlocker(&writeLock);

    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::open";
    readPos = 0;
    writePos = cacheFileWriteOnly.size();
    contentLength = writePos;
    bool b  = cacheFileWriteOnly.open(QIODevice::WriteOnly | QIODevice::Unbuffered);
    bool b2 = cacheFileReadOnly.open(QIODevice::ReadOnly | QIODevice::Unbuffered);
    if (DEBUG_IODEVICE)  qDebug() << "OPEN file: write"<<b<< "read"<<b2;
    return QIODevice::open(QIODevice::ReadWrite | QIODevice::Unbuffered) && b && b2;
}

bool MBbuffer::reset()
{
    QMutexLocker rlocker(&readLock);
    QMutexLocker wlocker(&writeLock);

    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::reset()";
    readPos = 0;
//    writePos= 0;
    cacheFileReadOnly.seek(0);
    QIODevice::seek(0);
//    cacheFileWriteOnly.seek(0);
//    chunkFile.seek(0);

    return true;
}

bool MBbuffer::seek( qint64 pos )
{
    if (pos>contentLength) return false;
    if (!reply.isNull()) reply->abort();

    QMutexLocker rlocker(&readLock);
    QMutexLocker wlocker(&writeLock);

    bool b1 = cacheFileReadOnly.seek(pos);
    bool b2 = QIODevice::seek(pos);
    bool b3 = cacheFileWriteOnly.seek(pos);
    if (b1) readPos = pos;
    if (b3) writePos = pos;
    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::seek" << "at pos" << pos << "*** "
             << "writeOnlyFile:" << cacheFileWriteOnly.size()
             << "readOnlyFile:" << cacheFileReadOnly.size()
             << "contentLenght:" << contentLength;
    return b1 && b2;
}

qint64 MBbuffer::size() const
{
    if (DEBUG_IODEVICE) qDebug() << "size()" << contentLength;
    return contentLength;
}

qint64 MBbuffer::bytesAvailable() const
{
    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::bytesAvailable()" << contentLength-readPos;
    QIODevice::bytesAvailable();
    return contentLength-readPos;
}

// For buffered devices, this function returns the number of bytes waiting
// to be written. For devices with no buffer, this function returns 0.
qint64 MBbuffer::bytesToWrite () {
    if (DEBUG_IODEVICE) qDebug() << "bytesToWrite";
    return 0;
}


qint64 MBbuffer::peek( char * data, qint64 maxSize )
{
    QMutexLocker locker(&readLock);

    if (DEBUG_IODEVICE) qDebug() << "MBbuffer::peek()";
    qint64 oldReadPos = readPos;
    qint64 n = readData(data, maxSize);
    readPos = oldReadPos;
    return n;
}

qint64 MBbuffer::read( char * data, qint64 maxSize )
{
    return readData(data, maxSize);
}

QByteArray MBbuffer::read( qint64 maxSize )
{
    QByteArray r;
    readData(r.data(), maxSize);
    return r;
}

qint64 MBbuffer::readData(char* data, qint64 maxSize)
{
    if (!cacheFileReadOnly.isOpen()) return 0;
    QMutexLocker rlocker(&readLock);

    if (maxSize<=0) return -1;
    // read what remaining if more data is requested
    if (readPos + maxSize > contentLength) {
        maxSize = contentLength - readPos;
        if (DEBUG_IODEVICE) qDebug() << "Read remaining=" << maxSize << "from a total of " << contentLength;
    }
    if (maxSize<=0) return -1;

    // if the 1st try to read data fails, the 2nd use getSync
    qint64 n;
        // Prevent to download many times first bytes when video starts (first BUFFER_SIZE must always been downloaded here)
//        if (readPos > BUFFER_SIZE) {
            if ( !readPosInRange(readPos) )
            {
                //        qDebug() << "MBbuffer::readData" << readPos << getAsyncThread->lowRangeValue() << getAsyncThread->currentWritingPos();
                getSync(readPos, readPos+maxSize);
            }
//        }

        // Read content
        cacheFileReadOnly.seek(readPos);
        n = cacheFileReadOnly.read(data, maxSize);
        if (n > 0) {
            if (DEBUG_IODEVICE) qDebug() << "READ" << maxSize << "at=" << readPos << "from a total of " << contentLength;
            readPos += n;
            emit readingData(readPos, maxSize);
        }
    return n;
}


qint64 MBbuffer::writeData(const char* data, qint64 maxSize)
{
    QMutexLocker locker(&writeLock);

    qint64 n = cacheFileWriteOnly.write(data, maxSize);
    if (n > 0) {
        emit dataWritten((qint64)writePos, (qint64)n);
        int percStart, percEnd;
        percStart = 1000LL * writePos / contentLength;
        percEnd   = 1000LL * (writePos+n) / contentLength;
        emit dataWritten(percStart, percEnd);
        if (DEBUG_IODEVICE)  qDebug() << "Written  " << n <<  "\t [" << writePos << "-" << (writePos+maxSize) << "] of " << contentLength;
        writePos += n;
        contentBytesWritten += n;
        emit bytesWritten(n);
        emit readyRead();
    }
    else
        if (DEBUG_IODEVICE) qDebug() << "Write failed" << cacheFileWriteOnly.error() << cacheFileWriteOnly.errorString() << "WRITEPOS" << writePos << "MAXSIZE" << maxSize;

    return n;
}

