/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

#ifndef MBBUFFER_H
#define MBBUFFER_H

#include <QIODevice>
#include <QThread>
#include <QTimer>
#include <QMutex>
#include <QFile>
#include <QEventLoop>
#include <QPointer>
#include <QThread>

#include <QUrl>
#include <QNetworkCookieJar>
#include <QNetworkCookie>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>
#include "mbcookiejar.h"

class MBbufferThread;
class MBbuffer;

class MBbufferThread : public QObject
{
    Q_OBJECT
public:
    MBbufferThread (bool asynchronous, QObject *parent = 0);
    ~MBbufferThread();
    void init(QUrl url, MBbuffer *device, qint64 rangeStart, qint64 rangeEnd, qint64 startAlreadyDownloaded);
    void run();
    bool NAMisRunning();
    qint64 lowerWrittenPos();
    qint64 highRangeValue();
    qint64 lowRangeValue();
    qint64 currentWritingPos();
    void resetBuffer();
    void setNAMcookies(MbCookieJar *cookieJar);

private:
    MBbuffer *buffer;
    QNetworkAccessManager *manager;
    QPointer<QNetworkReply>  reply;
    qint64 startPos, endPos, currWritePos, alreadyDownloadedPos;
    QUrl currentUrl;
    bool bufferReadyEmitted;
    bool async;
    MbCookieJar *cookies;
    MbCookieJar *tmpCookieJar; // Temporary cookie, the same as session cookie, but it will be deleted by QNetworkAccessManager

signals:
    void getFinished();
    void bufferReady();
    void bufferEmptyed();
    void getError();

public slots:
    void downloadProgress(qint64 current, qint64 total);
    void downloadFinished();
    void downloadError(QNetworkReply::NetworkError e);
    void abortNAM();
};

class MBbuffer : public QIODevice
{
    Q_OBJECT
public:
    explicit MBbuffer(QUrl url, MbCookieJar *cookieJar = 0, QObject *parent = 0);
    ~MBbuffer();
    void init(QUrl url);
    void setUrl(QUrl url);
    void getAsync(qint64 start=0, qint64 end=0);
    void getSync(qint64 start=0, qint64 end=0);
    qint64 getContentLenght();
    void startDownloadFromBytePos(qint64 fromPos);
    void startDownloadFromPerc(int perc);
    void seekWriteFile(qint64 pos);
    qint64 readPosValue();
    bool readPosInRange(qint64 pos);
    void setNAMcookies(MbCookieJar *cookieJar);
    bool isBuffering();
    static bool singleton;

    // Reimplemented QIODevice functions
    bool   atEnd() const;
    qint64 bytesAvailable() const;
    qint64 bytesToWrite();
    void   close();
    bool   isSequential() const;
    bool   open( OpenMode mode );
    qint64 peek(char * data, qint64 maxSize);
    qint64 read( char * data, qint64 maxSize );
    QByteArray read( qint64 maxSize );
    bool   reset();
    bool   seek( qint64 pos );
    qint64 size() const;

protected:
    // Reimplemented QIODevice functions
    qint64 readData(char* data, qint64 maxSize);
    qint64 writeData(const char* data, qint64 maxSize);

private:
//    QNetworkAccessManager *manager;
    MBbufferThread *getAsyncThread;
    QPointer<QNetworkReply>  reply;
    QFile  cacheFileWriteOnly;
    QFile  cacheFileReadOnly;
    qint64 contentLength;
    qint64 contentBytesWritten;
    qint64 readPos;
    qint64 writePos;
    mutable QMutex readLock;
    mutable QMutex writeLock;
    mutable QMutex getLock;
    QUrl currentUrl;
    qint64 downloadSize;
    QTimer bufferTimer;
    bool buffering;
    MbCookieJar *cookies;
    MbCookieJar *tmpCookieJar; // Temporary cookie, the same as session cookie, but it will be deleted by QNetworkAccessManager


    qint64 fillWithZeros(qint64 size);

signals:
    void dataWritten(int percStart, int percEnd);
    void dataWritten(qint64 pos, qint64 size);
    void totalSizeAcquired(qint64 size);
    void readingData(qint64 readPos, qint64 maxSize);
    void bufferReady();
    void bufferEmptyed();

public slots:
    void startDownload();
    void onBufferReady();
    void onBufferEmptyed();

};

#endif // MBBUFFER_H
