/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

#include <QDesktopWidget>
#include <QImage>
#include <QPen>
#include <QPainter>
#include <QDebug>
#include <QNetworkCookieJar>

#include "network.h"
#ifndef MOBILE_APP
#include "mainwindow.h"
#endif
#include "ui_network.h"

#ifdef MOBILE_APP
#define PROGRESS_HEIGHT 60
#else
#define PROGRESS_HEIGHT 15
#endif

///////////////////////////////////////////////////////////////////////////////////////
// To experiment downloading files on linux and not get stuf loaded instantly after
// a redownload the same file, do this as root:
// "sync; echo 3 > /proc/sys/vm/drop_caches"
// For example if I download a 50 Mb local file and then redownload it, also in an other
// session, I will get it instantly with no time to see what happens
///////////////////////////////////////////////////////////////////////////////////////

// If a parent is not given and if setPosition is not called,
// the progress will be shown at center of desktop
progressWidget::progressWidget(bool perc, QWidget *parent) :
        QWidget(parent,
                (Qt::FramelessWindowHint |
                Qt::WindowStaysOnTopHint |
                Qt::X11BypassWindowManagerHint) &
                ~Qt::WindowCloseButtonHint )
{
    setAutoFillBackground(false);
    setStyleSheet("background:transparent;background-color:transparent;");
    this->setAttribute(Qt::WA_TranslucentBackground,true);
//    this->setWindowFlags( ( (this->windowFlags() | Qt::CustomizeWindowHint)
//                              & ~Qt::WindowCloseButtonHint) );

    percent=0;
    display_percentage = perc;
    setDiameter(80);
    gradientColor=0x00FF00;

    if (parent==0) {
        setPosition(QApplication::desktop()->screenGeometry(-1).width() / 2,
                    QApplication::desktop()->screenGeometry(-1).height() / 2);
    } else {
        setPosition( parent->width() / 2,
                     parent->height() / 2 );
    }
}

progressWidget::~progressWidget()
{
}


void progressWidget::setProgressColor(int r, int g, int b)
{
    gradientColor=(r<<16)+(g<<8)+b;
    setDiameter(diameter);
}

void progressWidget::setDiameter(int d)
{
    diameter = d;
    half_diameter=d/2;

    gradient=QRadialGradient(half_diameter, half_diameter, // center
                             half_diameter,                // radius
                             half_diameter, half_diameter);// focal point
    gradient.setColorAt(0,    QColor::fromRgb(0, 0, 0, 0));
    gradient.setColorAt(0.2,  QColor::fromRgb(0, 0, 0, 0));
    gradient.setColorAt(0.75, QColor::fromRgb(gradientColor));
    gradient.setColorAt(0.85, QColor::fromRgb(0, 0, 0, 0));
    gradient.setColorAt(0.95, QColor::fromRgb(255,255,255, 255));
    gradient.setColorAt(1,    QColor::fromRgb(0, 0, 0, 255));
    brush=QBrush(gradient);
}

void progressWidget::setPosition(int x, int y)
{
    centerx = x;
    centery = y;
    this->setGeometry(centerx-half_diameter, centery-half_diameter, diameter, diameter);
}

// Set position at center of the given rect
void progressWidget::setPosition(QRect r)
{
    centerx = r.x()+r.width()/2;
    centery = r.y()+r.height()/2;
    this->setGeometry(centerx-half_diameter, centery-half_diameter, diameter, diameter);
}

void progressWidget::setProgress(int p)
{
    percent=p;
    this->update();
}

void progressWidget::setProgress(qint64 current, qint64 total)
{
    percent=current*100.0/total;
    this->update();
}

void progressWidget::paintEvent(QPaintEvent * event)
{
    Q_UNUSED(event);

    QPainterPath path;
    path.moveTo(half_diameter, half_diameter);
    path.arcTo(rect(), 0, 360*percent/100); // percent * 3.6
    path.closeSubpath();

    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing);
    painter.setBrush(brush);
    painter.setPen( QPen(QColor(0,0,0,0)) );
    painter.drawPath(path);

    // Write percent number
    if (display_percentage) {
        painter.setPen(QColor(255,255,255,255));
        painter.setFont(QFont("Helvetica [Cronyx]", 10, QFont::Bold));
        painter.drawText( QRectF(0,0,diameter,diameter), Qt::AlignHCenter | Qt::AlignVCenter, QString::number(percent)+"%" );
        painter.setPen(QColor(0,0,0,255));
        painter.setFont(QFont("Helvetica [Cronyx]", 9, QFont::Bold));
        painter.drawText( QRectF(0,0,diameter,diameter), Qt::AlignHCenter | Qt::AlignVCenter, QString::number(percent)+"%" );
    }
}
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
progressBar::progressBar(int height, QString text_label, int i, QWidget *parent) :
        QWidget(parent)
{
    ID=i;
    layout=new QHBoxLayout(this);

    PB_cancel = new QPushButton();
    PB_cancel->setIcon(QIcon(":/icons/icons/stop.png"));
    PB_cancel->setIconSize(QSize(height, height));
    PB_cancel->setMaximumSize(height, height);
    PB_cancel->setMinimumSize(height, height);
    connect(PB_cancel, SIGNAL(clicked()), this, SLOT(PB_progress_click()) );

    progress = new progressWidget(false);
    progress->setFixedSize(height, height);
    progress->setDiameter(height);
    progress->setPosition(height/2, height/2);

    L_percent = new QLabel();
    L_percent->setText("0%");
    L_percent->setMinimumHeight(height);
    L_percent->setMaximumHeight(height);
    L_percent->setMinimumWidth(100);
    L_percent->setMaximumWidth(100);
//    L_percent->adjustSize();

    L_text = new QLabel();
    L_text->setText(text_label);
    L_text->setMinimumHeight(height);
    L_text->setMaximumHeight(height);
    L_text->adjustSize();

    layout->setSpacing(5);
    layout->setMargin(0);
    layout->addWidget(PB_cancel);
    layout->addWidget(progress);
    layout->addWidget(L_percent);
    layout->addWidget(L_text);
}

progressBar::~progressBar()
{
    delete layout;
    delete L_percent;
    delete L_text;
    delete PB_cancel;
    delete progress;
}

void progressBar::setProgressError(QString error)
{
    L_text->setStyleSheet("background-color: qlineargradient(spread:pad, x1:0, y1:0, x2:1, y2:0, stop:0.313901 rgba(200, 120, 120, 255), stop:1 rgba(255, 255, 255, 120));");
    L_text->setToolTip(error);
}

void progressBar::setProgress(int percent)
{
    progress->setProgress(percent);
    L_percent->setText(QString::number(percent)+"%");
}

void progressBar::setProgress(qint64 current, qint64 total)
{
    progress->setProgress(current, total);
    L_percent->setText(QString::number(current/1024)+"/"+QString::number(total/1024)+" Kb");
}

void progressBar::PB_progress_click()
{
    emit on_PB_progress_button_clicked(ID);
}

/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
int network::max_concurrent_downlods=1;
bool network::outDebug=true;
//int network::n_current_downloads=0;
QPointer<MbCookieJar> network::cookieJar=0;

network::network(bool progressVisible, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::network)
{
    ui->setupUi(this);
    setAttribute(Qt::WA_DeleteOnClose);
#ifdef MOBILE_APP
    QFont f=this->font();
    f.setPointSize(15);
    this->setFont(f);
#endif

    manager=new QNetworkAccessManager(this);
    if (cookieJar.isNull())
        cookieJar = new MbCookieJar();

    if (tmpCookieJar.isNull())
        tmpCookieJar = new MbCookieJar();

    tmpCookieJar->copyFrom(cookieJar);
    manager->setCookieJar(tmpCookieJar);

    n_current_downloads=0;
    n_request=0;
    view_progress=progressVisible;
    if (view_progress) this->show();
    else this->hide();
    this->setFixedWidth(800);//MainWindow::pointerToMainWindow->geometry().width());
}

network::~network()
{
    int n;

    cookieJar->appendFrom((MbCookieJar*)manager->cookieJar());

    n=queue_list.size()-1;
    while (n>=0) {
        if (!queue_list[n].down.isNull()) delete queue_list[n].down;
        if (view_progress)
            if (!queue_list[n].progress.isNull()) delete queue_list[n].progress;
        n--;
    }
    queue_list.clear();
    delete manager;
    delete ui;
}

void network::remove_item(int i)
{
    // remove corresponding item with downloaded ID
    int n=0;
    while (n<queue_list.size()) {
        if (queue_list.at(n).ID == i) break;
        n++;
    }
    if ( n<queue_list.size() )
    {
        if (queue_list[n].error.isEmpty()) {
            delete queue_list[n].down;
            if (view_progress) delete queue_list[n].progress;
        }
        else {
                if (view_progress) queue_list[n].progress->setProgressError(queue_list[n].error);
        }
        queue_list.removeAt(n);
        n_current_downloads--;
    }
    ui->L_info->setText(tr("Remaining downloads: ")+QString::number(queue_list.size()));
}

void network::check_queue()
{
    if (n_current_downloads >= max_concurrent_downlods)
        return;

    // If there are downloads in queue use those
    if (queue_list.size()>0) {
        int n=0;
        // Find first available download item
        while (n<queue_list.size()) {
            if (!queue_list[n].downloading) break;
            n++;
        }


        // if an item to download is found
        if (n<queue_list.size()) {
            n_current_downloads++;
            queue_list[n].down=new download(manager, this);
            connect(queue_list[n].down, SIGNAL(downloadSlotError(int,QNetworkReply::NetworkError)),
                    this, SLOT(get_error(int,QNetworkReply::NetworkError)) );
            connect(queue_list[n].down, SIGNAL(downloadFinished(int,QByteArray)),
                    this, SLOT(get_completed(int,QByteArray)) );
            connect(queue_list[n].down, SIGNAL(downloadProgress(int,qint64,qint64)),
                    this, SLOT(get_progress(int,qint64,qint64)) );

            if (view_progress) {
                queue_list[n].progress=new progressBar( PROGRESS_HEIGHT, queue_list[n].url.toString(), queue_list[n].ID, this);
                queue_list[n].progress->progress->setProgressColor(0,0,255);
                connect(queue_list[n].down, SIGNAL(downloadProgressPercent(int)),
                        queue_list[n].progress, SLOT(setProgress(int)) );
                connect(queue_list[n].progress, SIGNAL(on_PB_progress_button_clicked(int)),
                        this, SLOT(onPB_progress_cancel_clicked(int)) );
                ui->verticalLayout_2->insertWidget(0, queue_list[n].progress);
                ui->L_info->setText(tr("Remaining downloads: ")+QString::number(queue_list.size()));
            }

            queue_list[n].downloading=true;
            queue_list[n].down->get( queue_list[n].ID, queue_list[n].url );
        }
    }
}

// return true if there where downloads
bool network::cancellAllDownloads()
{
    bool b=false;
    int n;
    n=queue_list.size();
    if (n>0) b=true;
    while (n>0) {
        delete queue_list[0].down;
        if (view_progress) delete queue_list[0].progress;
        queue_list.removeAt(0);
        n--;
    }
    return b;
}

void network::get(QString url)
{
    net_item n;

    n.ID  = n_request;
    n.url = url;
    n.downloading=false;
    queue_list << n;
    n_request++;

    check_queue();
}

void network::get_completed(const int i, const QByteArray &data)
{
    emit downloadFinished(i, data);
    remove_item(i);
    if (queue_list.size()==0)
        emit queue_completed();
    else
        check_queue();

    cookieJar->appendFrom((MbCookieJar*)manager->cookieJar());
}

void network::get_error(int i, QNetworkReply::NetworkError e)
{
    int n=0;
    while (n<queue_list.size()) {
        if (queue_list[n].ID==i) break;
        n++;
    }
    queue_list[n].error=queue_list[n].down->errorHumanString;

    emit downloadError(i, e);
    remove_item(i);
    if (queue_list.size()==0)
        emit queue_completed();
    else
        check_queue();
}

void network::get_progress(int i, qint64 current, qint64 total)
{
    emit downloadProgress(i, current, total);
//    emit stillDownloading();
}

void network::onPB_progress_cancel_clicked(int i)
{
    int n=0;
    while (n<queue_list.size()) {
        if (queue_list.at(n).ID == i) break;
        n++;
    }
    if (n<queue_list.size()) {
        if (!queue_list[n].down.isNull()) delete queue_list[n].down;
        if (view_progress) delete queue_list[n].progress;
        queue_list.removeAt(n);
    }
}
/////////////////////////////////////////////////
// Get synchroneous download
QByteArray network::get_synchronous(QString url, bool showSingleProgress, QRect position)
{
    QByteArray b="";
    QString s;
    progressBar *progressDialog=0;
    QEventLoop q;

    s="Downloading: "+url;
    ui->L_info->setText(s);
//    qDebug() << s;

//    QWidget *activeWidget=qApp->activeWindow();
//    QCoreApplication::postEvent(activeWidget, new networkStartedEvent);

    show_single_progress_widget=showSingleProgress;

    gettSynchronousCancelled=false;

//    QNetworkRequest request(QUrl(url.toUtf8()));

//    request.setRawHeader("User-Agent", "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b11) Gecko/20100101 Firefox/4.0b11");
//    request.setRawHeader("Accept-Charset", "ISO-8859-1,utf-8;q=0.7,*;q=0.7");
//    request.setRawHeader("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
//    request.setRawHeader("Accept-Encoding", "gzip, deflate");
//    request.setRawHeader("Accept-Language", "en-us,en;q=0.5");
//    request.setRawHeader("Connection", "Keep-Alive");

//    reply_synchronous = manager->get(request);

    tmpCookieJar->copyFrom(cookieJar);
    manager->setCookieJar(tmpCookieJar);


    reply_synchronous = manager->get(QNetworkRequest(QUrl(url)));

    if (view_progress)
    {
        progressDialog=new progressBar( PROGRESS_HEIGHT, url, 0, this);
        progressDialog->progress->setProgressColor(255,0,0);
        progressDialog->progress->show();
        connect(reply_synchronous, SIGNAL(downloadProgress(qint64,qint64)),
                progressDialog, SLOT(setProgress(qint64,qint64)) );
        connect(this->ui->PB_cancel, SIGNAL(clicked()),
                this, SLOT(onGetSynchronousCancelled()) );
        connect(this->ui->PB_cancel, SIGNAL(clicked()),
                &q, SLOT(quit()) );
        ui->verticalLayout_2->insertWidget(0, progressDialog);
    }


    if (showSingleProgress) {
#ifndef MOBILE_APP
        progress = new progressWidget(true, MainWindow::pointerToMainWindow);
#else
        progress = new progressWidget(true, 0);
#endif
        if (position.width()!=0) progress->setPosition(position);
        connect(reply_synchronous, SIGNAL(downloadProgress(qint64,qint64)),
                progress, SLOT(setProgress(qint64,qint64)) );
        progress->show();
    }


//    connect(manager, SIGNAL(finished(QNetworkReply*)), &q, SLOT(quit()));
    connect(reply_synchronous, SIGNAL(finished()), &q, SLOT(quit()) );
//    connect(reply_synchronous, SIGNAL(error(QNetworkReply::NetworkError)), &q, SLOT(quit()) );

    q.exec();


    cookieJar->appendFrom((MbCookieJar*)manager->cookieJar());


//    bool success = reply_synchronous->isReadable();

    if (!gettSynchronousCancelled) {
        QString forwardedUrl;
        forwardedUrl= reply_synchronous->attribute(QNetworkRequest::RedirectionTargetAttribute).toUrl().toString();
        // It's possible that with no error a forwarded link is been returned
        // like for deviantArt images, so we check for this
        if (!forwardedUrl.isEmpty()) {
            s="Forwarded to: "+forwardedUrl;
            ui->L_info->setText(s);
            qDebug() << s ;

            reply_synchronous = manager->get(QNetworkRequest(QUrl(forwardedUrl)));
            connect(reply_synchronous, SIGNAL(finished()), &q, SLOT(quit()) );
            connect(reply_synchronous, SIGNAL(error(QNetworkReply::NetworkError)), &q, SLOT(quit()) );
            if (showSingleProgress)
                connect(reply_synchronous, SIGNAL(downloadProgress(qint64,qint64)),
                        progress, SLOT(setProgress(qint64,qint64)) );
            if (view_progress) {
                progressDialog->progress->setProgressColor(255,0,0);
                progressDialog->progress->show();
                connect(reply_synchronous, SIGNAL(downloadProgress(qint64,qint64)),
                        progressDialog, SLOT(setProgress(qint64,qint64)) );
            }
            q.exec();

            cookieJar->appendFrom((MbCookieJar*)manager->cookieJar());
        }

        b=reply_synchronous->readAll();
        if (reply_synchronous->error()!=QNetworkReply::NoError) {
            qDebug() << "Error: " << reply_synchronous->error() << reply_synchronous->errorString();
        }
        else
            emit queue_completed();
    }

    show_single_progress_widget = false;
    if (showSingleProgress) delete progress;
    if (view_progress) delete progressDialog;
//    QCoreApplication::postEvent(activeWidget, new networkEndedEvent);
    return b;
}

void network::onGetSynchronousCancelled()
{
    gettSynchronousCancelled=true;
}

////////////////////////////////////
void network::setProgressColor(int ID, int r, int g, int b)
{
    if (ID>queue_list.size()) return;
    queue_list[ID].progress->progress->setProgressColor(r,g,b);
}

////////////////////////////////////
void network::on_PB_cancel_clicked()
{
    this->close();
}

////////////////////////////////////
// Avoid progressWidget to lose staysOnTopHint. For example when mediaplayer switch full screen
void network::onSetProgressWidgetOnTop()
{
    if (show_single_progress_widget) {
        delete progress;
        progress = new progressWidget(true, this);
        connect(reply_synchronous, SIGNAL(downloadProgress(qint64,qint64)),
                progress, SLOT(setProgress(qint64,qint64)) );
        progress->show();
        qDebug() << "onSetProgressWidgetOnTop";
    }
}

/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
download::download( QNetworkAccessManager *m_na, QObject *parent )
{
    Q_UNUSED(parent)
    errorHumanString="";
    useCacheFile=false;

    // If m_na is not passed, we assume we want to use a stremBuffer file
    if (m_na!=0) {
        manager = m_na;
    }
    else {
        manager = new QNetworkAccessManager(this);
//        useCacheFile=true;
    }
}

download::~download()
{
    if (useCacheFile) {
//        buffer->cacheFile.remove();
    }
#if QT_VERSION > 0x040600
    if (reply->isRunning()) reply->abort();
#else
    if (reply->isOpen()) reply->abort();
#endif
}

// Set the global session cookies acquired by network class during the whole life of the application
void download::setGlobalCookies()
{
    if (network::cookieJar.isNull())
        return;

    if (tmpCookieJar.isNull())
        tmpCookieJar = new MbCookieJar();

    tmpCookieJar->copyFrom(network::cookieJar);
    manager->setCookieJar(tmpCookieJar);
}

void download::get(int n, QUrl url)
{
    ID=n;
    get_url = url;

    //    QNetworkRequest request(QUrl(url.toUtf8()));

    //    request.setRawHeader("User-Agent", "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b11) Gecko/20100101 Firefox/4.0b11");
    //    request.setRawHeader("Accept-Charset", "ISO-8859-1,utf-8;q=0.7,*;q=0.7");
    //    request.setRawHeader("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
    //    request.setRawHeader("Accept-Encoding", "gzip, deflate");
    //    request.setRawHeader("Accept-Language", "en-us,en;q=0.5");
    //    request.setRawHeader("Connection", "Keep-Alive");

    //    reply_synchronous = manager->get(request);

    reply = manager->get(QNetworkRequest(url));
    connect(reply, SIGNAL(error(QNetworkReply::NetworkError)),
            this, SLOT(onSlotError(QNetworkReply::NetworkError)) );
    connect(reply, SIGNAL(finished()),
            this, SLOT(onDownloadFinished()) );
    connect(reply, SIGNAL(downloadProgress(qint64,qint64)),
            this, SLOT(onProgressChanged(qint64,qint64)) );
}

void download::onDownloadFinished()
{
    QString forwardedUrl;
    // It's possible that with no error a forwarded link is been returned
    // like for deviantArt images, so we check for this
    forwardedUrl= reply->attribute(QNetworkRequest::RedirectionTargetAttribute).toUrl().toString();
    if (!forwardedUrl.isEmpty())
    {
        qDebug() << "Forwarded to: "+forwardedUrl;
        get(ID, forwardedUrl);
        return;
    }
    QByteArray b;
    b=reply->readAll();
    emit downloadFinished(ID,b);
}

void download::onSlotError(QNetworkReply::NetworkError e)
{
    errorHumanString=reply->errorString();
    reply->disconnect(this);
    qDebug() << "download error: " << e << errorHumanString;
    emit this->downloadSlotError(ID,e);
}

void download::onProgressChanged(qint64 current, qint64 total)
{
    emit downloadProgress(ID, current, total);
    emit downloadProgressPercent(current*100.0/total);
}

//void download::onReadyRead()
//{
//    if (!reply) // event sender deleted!
//        return;

//    // read available bytes from http stream and write them to cache file
//    qint64 n = reply->bytesAvailable();
//    //qDebug() << "Reading " << n << " bytes";
//    n = buffer->write(reply->read(n));
////    qDebug() << "Wrote " << n;
//}
