/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

#ifndef NETWORK_H
#define NETWORK_H

#include <QDialog>
#include <QLabel>
#include <QPushButton>
#include <QHBoxLayout>
#include <QList>
#include <QNetworkCookieJar>
#include <QNetworkCookie>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QPointer>
#include <QEvent>
#include <QDebug>

#include "mbcookiejar.h"

namespace Ui {
    class network;
}

class progressWidget;
class progressBar;
class network;
class download;


class progressWidget : public QWidget
{
    Q_OBJECT

public:
    progressWidget(bool perc=false, QWidget *parent = 0);
    ~progressWidget();
    QRgb gradientColor;
    void setProgressColor(int r, int g, int b);
    void setDiameter(int d);
    void setPosition(int x, int y);
    void setPosition(QRect r);
    int percent;

protected:
    virtual void paintEvent ( QPaintEvent * event );

private:
    QRadialGradient gradient;
    QBrush brush;
    int half_diameter;
    int diameter;
    int centerx, centery;
    bool display_percentage;

public slots:
    void setProgress(int p);
    void setProgress(qint64 current, qint64 total);
};
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
class progressBar : public QWidget
{
    Q_OBJECT

public:
    progressBar(int height, QString text_label, int i, QWidget *parent = 0);
    ~progressBar();

    void setProgressError(QString error);

    progressWidget *progress;
    QPushButton *PB_cancel;
    QLabel *L_percent;
    QLabel *L_text;
    QHBoxLayout *layout;
    int ID;

public slots:
    void setProgress(int percent);
    void setProgress(qint64 current, qint64 total);
    void PB_progress_click();

signals:
    void on_PB_progress_button_clicked(int i);
};

///////////////////////////////////////////////////////
// send event while downloading ///////////////////////
enum { NETWOTK_ACTIVITY_STARTED = QEvent::User,
       NETWOTK_ACTIVITY_ENDED
     };

class networkStartedEvent : public QEvent
{
public:
  networkStartedEvent( )
      : QEvent( static_cast<QEvent::Type>(NETWOTK_ACTIVITY_STARTED) ) {};
};
class networkEndedEvent : public QEvent
{
public:
  networkEndedEvent( )
      : QEvent( static_cast<QEvent::Type>(NETWOTK_ACTIVITY_ENDED) ) {};
};

/////////////////////////////////////////////////////////
//  network queue manager  //////////////////////////////
class network : public QDialog {
    Q_OBJECT

public:

    struct net_item {
        int ID;
        QUrl url;
        QPointer<download> down;
        bool downloading;
        QPointer<progressBar> progress;
        QString error;
    } ;

    network(bool progressVisible=false, QWidget *parent = 0);
    ~network();

    Ui::network *ui;
    static int max_concurrent_downlods;
    static bool outDebug;
    static QPointer<MbCookieJar> cookieJar; // Session cookies
    QPointer<MbCookieJar> tmpCookieJar;     // Temporary cookie, the same as session cookie, but it will be deleted by QNetworkAccessManager
    int n_request;
    /*static */int n_current_downloads;

    void get(QString url);
    bool cancellAllDownloads();
    QByteArray get_synchronous(QString url, bool showSingleProgress, QRect position=QRect(0,0,0,0));
    void setProgressColor(int ID, int r, int g, int b);
    QNetworkReply *reply_synchronous;

private:

    void check_queue();
    void remove_item(int i);

    QList<net_item> queue_list;
    QNetworkAccessManager *manager;
    bool view_progress;
    bool show_single_progress_widget;
    bool gettSynchronousCancelled;
    progressWidget *progress;

public slots:
    void get_completed(const int i, const QByteArray &data);
    void get_error(int i, QNetworkReply::NetworkError e);
    void get_progress(int i, qint64 current, qint64 total);
    void on_PB_cancel_clicked();
    void onPB_progress_cancel_clicked(int i);
    void onGetSynchronousCancelled();
    void onSetProgressWidgetOnTop();

signals:
    void downloadFinished(int i, QByteArray buffer);
    void downloadError(int i,QNetworkReply::NetworkError e);
    void downloadProgress(int i, qint64 current, qint64 total);
    void queue_completed();
    void stillDownloading();

};
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
class download : public QObject
{
    Q_OBJECT
public:
    download( QNetworkAccessManager *m_na = 0, QObject *parent = 0);
    ~download();
    void get(int n, QUrl url);
    void setGlobalCookies();

    QString errorHumanString;
    QNetworkReply *reply;
    QNetworkAccessManager *manager;

private:
    int ID;
    QUrl get_url;
    bool useCacheFile;
    QPointer<MbCookieJar> tmpCookieJar;

public slots:
    void onDownloadFinished();
    void onSlotError(QNetworkReply::NetworkError e);
    void onProgressChanged(qint64 current, qint64 total);
//    void onReadyRead();

signals:
    void downloadFinished(int i, QByteArray buffer);
    void downloadSlotError(int i,QNetworkReply::NetworkError e);
    void downloadProgress(int i, qint64 current, qint64 total);
    void downloadProgressPercent(int percent);
};

#endif // NETWORK_H
