/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

#include <QFile>
#include <QString>
#include <QBuffer>
#include <QMessageBox>
#include <QDebug>
#include "defines.h"
#include "saverestore.h"
#include "imageviewer2d.h"
#include "imagecontainer.h"
#include "mainwindow.h"

saverestore::saverestore(int op, ImageViewer2d *imagev, QString fn, QWidget *parent) :
    QWidget(parent)
{
    imageViewer=imagev;
    option=op;
    fileName=fn;

    switch (option) {
    case SAVE_AND_APPEND:
    case SAVE_TO_NEW_FILE:
        save();
        break;
    case RESTORE_AND_APPEND:
    case RESTORE_AND_REPLACE:
        restore();
        break;
    }
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//  FILE FORMAT
//
//        2 bytes of text for number of items
//          |
//          |
//          Item 1
//             !
//             4 BYTES for pixmap lenght
//             item pixmap
//             item data
//          !
//          !
//          Item 2
//          |
//          ...


void saverestore::save()
{
    QFile file(fileName);

    if (option==SAVE_TO_NEW_FILE)
    {
        if (file.exists())
        {
            if (QMessageBox::question(MainWindow::pointerToMainWindow,
                                      tr("File exist"),
                                      tr("Would you like to overwrite it ?"),
                                      QMessageBox::Ok | QMessageBox::Cancel)
                == QMessageBox::Cancel) return;
        }
        if (!file.open(QIODevice::WriteOnly))
        {
            QMessageBox::warning(MainWindow::pointerToMainWindow,
                                 tr("error"),
                                 tr("The file couldn't be opened"),
                                 QMessageBox::Ok);
            return;
        }
    }
    if (option==SAVE_AND_APPEND)
        if (!file.open(QIODevice::Append)) {
//            file.seek(file.size());
            QMessageBox::warning(MainWindow::pointerToMainWindow,
                                 tr("error"),
                                 tr("The file couldn't be opened"),
                                 QMessageBox::Ok);
            return;
        }

    QByteArray bytes;
    QBuffer buffer;
    QDataStream out(&file);
    imageContainer *item;
    int i;

    out.setVersion(QDataStream::Qt_4_5);
//    out.setByteOrder(QDataStream::LittleEndian);
    // Write 2 bytes for the number of items
    if (option==SAVE_TO_NEW_FILE)
        out << (quint16) imageViewer->itemsList.size();

    foreach (item , imageViewer->itemsList) {

        bytes.clear();
        buffer.setBuffer(&bytes);
        item->p.save(&buffer, "JPG");
        out << (quint32)buffer.size();

        out.writeRawData(buffer.data(), buffer.size());
        buffer.close();

        // WRITE STRUCT DATA
        out << (quint8) item->image.kind;
        out << item->image.description_row1;
        out << item->image.description_row2;
        out << item->image.description_row3;
        out << item->image.image_tburl;
        out << item->image.image_url;
        out << item->image.image_source;
            // write GoogleImageStruct
            out << item->image.googleImageInfo.width;
            out << item->image.googleImageInfo.height;
            out << item->image.googleImageInfo.imageId;
            out << item->image.googleImageInfo.tbWidth;
            out << item->image.googleImageInfo.tbHeight;
            out << item->image.googleImageInfo.unescapedUrl;
            out << item->image.googleImageInfo.url;
            out << item->image.googleImageInfo.visibleUrl;
            out << item->image.googleImageInfo.title;
            out << item->image.googleImageInfo.titleNoFormatting;
            out << item->image.googleImageInfo.originalContextUrl;
            out << item->image.googleImageInfo.content;
            out << item->image.googleImageInfo.contentNoFormatting;
            out << item->image.googleImageInfo.tbUrl;
            out << item->image.googleImageInfo.descr;
            out << item->image.googleImageInfo.kind;
            out << item->image.googleImageInfo.size;
            // write YouTubeStruct
            //   don't write video_info because are only temporary urls
            out << item->image.youTubeInfo.id;
            out << item->image.youTubeInfo.video_id;
            out << item->image.youTubeInfo.published;
            out << item->image.youTubeInfo.updated;
            out << item->image.youTubeInfo.title;
            out << item->image.youTubeInfo.content;
            out << item->image.youTubeInfo.author_name;
            out << item->image.youTubeInfo.author_uri;
            for (i=0; i<4; i++)
            {
                out << item->image.youTubeInfo.p[i].width;
                out << item->image.youTubeInfo.p[i].height;
                out << item->image.youTubeInfo.p[i].url;
            }
            out << item->image.youTubeInfo.mediaUrl;
            out << item->image.youTubeInfo.duration;
            out << item->image.youTubeInfo.rating_average;
            out << item->image.youTubeInfo.rating_numRaters;
            out << item->image.youTubeInfo.statistics_viewCount;
            out << (quint8) item->image.youTubeInfo.ID_preferred_video_format;
    }
    file.flush();
    file.close();
    qDebug() << "Wrote " << imageViewer->itemsList.size() << " items";

    // update number of items in the first 2 byte of the file
    // (maybe I choosen the long way to do this :-) )
    if (option==SAVE_AND_APPEND) {
        quint16 old_number,new_number;
        file.open(QIODevice::ReadWrite);
        out >> old_number;
        file.seek(0);
        uchar *first_2_bytes = file.map(0, 2);

        new_number = (quint16) (old_number+imageViewer->images.size());
        first_2_bytes[0] = (uchar) (new_number >> 8);
        first_2_bytes[1] = (uchar) (new_number);

        file.unmap(first_2_bytes);
        file.flush();
        file.close();
    }
}

void saverestore::restore()
{
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly))
    {
        QMessageBox::warning(MainWindow::pointerToMainWindow,
                             tr("error"),
                             tr("The file couldn't be opened"),
                             QMessageBox::Ok);
        return;
    }

    QDataStream in(&file);
    QByteArray pixmapData;
    char *s;

    in.setVersion(QDataStream::Qt_4_5);
    if (option==RESTORE_AND_REPLACE) imageViewer->clear_scene();

    quint16 n_items;
    quint32 pixmap_size;
    imageStruct is;
    quint8 kind;
    int i;
    int start_item_ID;

    start_item_ID = imageViewer->itemsList.size();
    in >> n_items;
    for (int n=0; n< n_items; n++) {
        // Read pixmap bytes lenght
        in >> pixmap_size;
        // Read pixmap data
        s=new char[pixmap_size+1];
        in.readRawData(s, pixmap_size);
        pixmapData.clear();
        pixmapData.append(s,pixmap_size);
        delete s;

        // READ ITEM DATA
       in >> kind;
       is.kind = (imageContainerKind)kind;
       in >> is.description_row1;
       in >> is.description_row2;
       in >> is.description_row3;
       in >> is.image_tburl;
       in >> is.image_url;
       in >> is.image_source;
          // read googleImageInfo
          in >> is.googleImageInfo.width;
          in >> is.googleImageInfo.height;
          in >> is.googleImageInfo.imageId;
          in >> is.googleImageInfo.tbWidth;
          in >> is.googleImageInfo.tbHeight;
          in >> is.googleImageInfo.unescapedUrl;
          in >> is.googleImageInfo.url;
          in >> is.googleImageInfo.visibleUrl;
          in >> is.googleImageInfo.title;
          in >> is.googleImageInfo.titleNoFormatting;
          in >> is.googleImageInfo.originalContextUrl;
          in >> is.googleImageInfo.content;
          in >> is.googleImageInfo.contentNoFormatting;
          in >> is.googleImageInfo.tbUrl;
          in >> is.googleImageInfo.descr;
          in >> is.googleImageInfo.kind;
          in >> is.googleImageInfo.size;
          // read youTubeInfo
          in >> is.youTubeInfo.id;
          in >> is.youTubeInfo.video_id;
          in >> is.youTubeInfo.published;
          in >> is.youTubeInfo.updated;
          in >> is.youTubeInfo.title;
          in >> is.youTubeInfo.content;
          in >> is.youTubeInfo.author_name;
          in >> is.youTubeInfo.author_uri;
          for (i=0; i<4; i++)
          {
              in >> is.youTubeInfo.p[i].width;
              in >> is.youTubeInfo.p[i].height;
              in >> is.youTubeInfo.p[i].url;
          }
          in >> is.youTubeInfo.mediaUrl;
          in >> is.youTubeInfo.duration;
          in >> is.youTubeInfo.rating_average;
          in >> is.youTubeInfo.rating_numRaters;
          in >> is.youTubeInfo.statistics_viewCount;
          in >> is.youTubeInfo.ID_preferred_video_format;

       is.ID = n+start_item_ID;

       imageViewer->sceneOGL.addItem( imageViewer->add_imageContainer(pixmapData, is) );
    }
    file.close();
    imageViewer->arrange_items(false);
}
