/***************************************************************************
*  Media Downloader. Search and download as many images or videos that     *
*    fit your search criteria                                              *
*  Copyright (C) 2009-2011  Marco Bavagnoli - lil.deimos@gmail.com         *
*                                                                          *
*  This program is free software: you can redistribute it and/or modify    *
*  it under the terms of the GNU General Public License as published by    *
*  the Free Software Foundation, either version 3 of the License, or       *
*  (at your option) any later version.                                     *
*                                                                          *
*  This program is distributed in the hope that it will be useful,         *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*  GNU General Public License for more details.                            *
*                                                                          *
*  You should have received a copy of the GNU General Public License       *
*  along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
***************************************************************************/

#include <QDebug>
#include <QDomDocument>
#include <QList>
#include <QRegExp>
#include <QMessageBox>
#include "youtube.h"

QString unescapeCharacters(QString s)
{
    QRegExp rx;
    QChar c;
    int pos;
    bool ok;

    rx.setMinimal(true);
    rx.setPattern("(%[A-F|0-9].)");
    pos=0;
    while ((pos = rx.indexIn(s, pos)) != -1)
    {
        c=rx.cap(1).toInt(&ok, 16);
        s=s.replace(rx.cap(1), c);
    }
    return s;
}

QString getVideoID(YouTubeStruct &tube)
{
    QRegExp rx;

    // if is not set the 11 chars video id
    if (tube.video_id.isEmpty()) {
        // if also tube.id is empy nothing to do
        if (tube.id.isEmpty()) return 0;
        // Retrieve video id from url (QString id)
        rx.setPattern(".*videos/(.*)");
        rx.setMinimal(false);
        rx.indexIn(tube.id);
        if (rx.numCaptures()<1) {
            QMessageBox::critical ( 0, "Error", "Error in the retrieved url:\n"+tube.id);
            return "";
        }
        return rx.cap(1);
    }
    return tube.video_id;
}

// http://en.wikipedia.org/wiki/YouTube#Quality_and_codecs
const video_formats youTube::videoFormats[13] = {
    {45,"1280x720","WebM",1280*720},
    {44,"854x480","WebM", 854*480},
    {43,"640x360","WebM", 640*360},
    {37,"1920x1080","MP4",1920*1080},
    {22,"1280x720","MP4", 1280*720},
    {35,"854x480","FLV",  854*480},
    {34,"640x480","FLV",  640*480},
    {0, "640x480","FLV",  640*480},
    {5, "640x480","FLV",  640*480},
    {18,"480x360","MP4",  480*360},
    {6, "480x360","FLV",  480*360},
    {13,"176x144","3GP",  176*144},
    {17,"176x144","3GP",  176*144},
};

int youTube::ID_preferred_video_format=0; // is the index number of const video_formats youTube::videoFormats
bool youTube::only_mp4=true;
QString youTube::currentSearchText="";

youTube::youTube(QObject *parent) : QObject(parent)
{}

/////////////////////////////////////////////////////////////////////////////
// search into &tube an existing resolution less or
// equal than videoFormats[ID_preferred_video_format] and
// return the index of QList<youTubeVideoInfo> video_info to play or download
int youTube::getPreferredResolution(YouTubeStruct &tube, int ID)
{
    bool b=false;
    int i;
    int found=-1;

    qDebug() << "Preferred ID " << ID << " (" << videoFormats[ID].res << "-" << videoFormats[ID].container << ")";
    i=0;
    while (i<tube.video_info.size() && !b) {
        if (videoFormats[ID].fmt == tube.video_info[i].fmt.toInt())
        {
            if (only_mp4) {
                if (videoFormats[ID].container=="MP4") {
                    found=i;
                    b=true;
                }
            }
            else {
                found=i;
                b=true;
            }
        }
        i++;
    }
    // If no match resolution try to get a lower one
    if (!b) {
        long delta=1920*1080;
        long currentDelta=delta;
        i=0;
        while (i<tube.video_info.size()) {
            // if the difference of current video area and preferred res is less or equal then delta
            if (abs(tube.video_info[i].xres*tube.video_info[i].yres - videoFormats[ID].area) <= currentDelta)
            {
                delta=abs(tube.video_info[i].xres*tube.video_info[i].yres - videoFormats[ID].area);
                if (only_mp4) {
                    if (tube.video_info[i].container=="MP4") {
                        // Choose the lesser currentDelta
                        if (delta < currentDelta) {
                            found = i;
                            currentDelta = delta;
                        }
                    }
                }
                else {
                    // Choose the lesser currentDelta
                    if (delta < currentDelta) {
                        found = i;
                        currentDelta = delta;
                    }
                }
            }
            i++;
        }
    }
    // I surrender. Give the 1st in the list
    if (found==-1) found = 0;
    qDebug() << "Preferred resolution found: " << tube.video_info[found].fmt.toInt() <<
                " - " << tube.video_info[found].humanFmt <<
                " - " << tube.video_info[found].container << " index:" << found;
    return found;
}

/////////////////////////////////////////////////////////////////////////////
// Retrieve sychronously a list of video formats available to be played or downloaded
// from tube.id
int youTube::getTubeVideoInfo(YouTubeStruct &tube)
{
    if (tube.video_info.size()>0) return tube.video_info.size(); // If info are already filled return;

    QString videoID;
    QRegExp rx;

    videoID = getVideoID(tube);
    if (videoID.isEmpty()) return 0;

    // Gather video information and output error if fails
    network *net;
    QByteArray data;
    net = new network(false);
    data=net->get_synchronous("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=embedded&ps=default&eurl=&gl=US&hl=en",false);
    // If fails retry with different &el value
    if (data.isEmpty() || data.startsWith("status=fail"))
        data=net->get_synchronous("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=detailpage&ps=default&eurl=&gl=US&hl=en",false);
    if (data.isEmpty() || data.startsWith("status=fail"))
        data=net->get_synchronous("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=vevo&ps=default&eurl=&gl=US&hl=en",false);

//    qDebug() << data;
    if (data.isEmpty()) {
        QMessageBox::critical ( 0, "Error", "Error gathering video information");
        return 0;
    }
    if (data.startsWith("status=fail")) {
        rx.setPattern("reason=(.*)%");
        rx.setMinimal(true);
        rx.indexIn(data);
        QMessageBox::critical ( 0, "Error", "Error: "+rx.cap(1).replace("+"," "));
        return 0;
    }
    net->deleteLater();

    return getTubeVideoInfoFromData(tube, data);
}
/////////////////////////////////////////////////////////////////////////////
// Retrieve Asychronously a list of video formats.
// gotTubeVideoInfo(m_tube, i); emitted if retrieved else errorGettingInfo();
void youTube::getTubeVideoInfoAsynch(YouTubeStruct tube)
{
    videoID = getVideoID(tube);
    if (videoID.isEmpty()) return;
    n_try=1;
    m_tube=tube;
    net = new network(false);
    connect(net, SIGNAL(downloadFinished(int,QByteArray)), this, SLOT(onDownloadFinished(int,QByteArray)) );
    connect(net, SIGNAL(downloadError(int,QNetworkReply::NetworkError)), this, SLOT(onDownloadError(int,QNetworkReply::NetworkError)) );
    // Try 1st &el
    net->get("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=embedded&ps=default&eurl=&gl=US&hl=en");
}
void youTube::onDownloadFinished(int n, QByteArray data)
{
    Q_UNUSED(n)

    if (data.startsWith("status=fail")) {
        onDownloadError(0, QNetworkReply::NoError);
        return;
    }
    int i = getTubeVideoInfoFromData(m_tube, data);
    emit gotTubeVideoInfo(m_tube, i);
    net->deleteLater();
}
void youTube::onDownloadError(int n, QNetworkReply::NetworkError e)
{
    Q_UNUSED(n)
    n_try++;
    if (n_try==2 && e==QNetworkReply::NoError) // Try 2nd &el
    {
        net->get("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=detailpage&ps=default&eurl=&gl=US&hl=en");
        return;
    }
    if (n_try==3 && e==QNetworkReply::NoError) // Try 3th &el
    {
        net->get("http://www.youtube.com/get_video_info?&video_id="+videoID+"&el=vevo&ps=default&eurl=&gl=US&hl=en");
        return;
    }
    if (n_try > 3) emit errorGettingInfo();
    net->deleteLater();
}

////////////////////////////////////////////////////////////////////////
// giving retrieved data will return a filled tube.video_info and number of occurrences
int youTube::getTubeVideoInfoFromData(YouTubeStruct &tube, QByteArray data)
{
    QByteArray data_percented;
    youTubeVideoInfo mode;
    QRegExp rx;
    int count=0;

    data_percented=data;
    while (data.contains('%'))
        data = QByteArray().fromPercentEncoding(data);
//    qDebug() << data;

    // Retrieve video token
    rx.setPattern("&token=(.*)&");
    rx.setMinimal(true);
    rx.indexIn(data);
    if (rx.numCaptures()<1) {
        QMessageBox::critical ( 0, "Error", "Error gathering video token information");
        return count;
    }
    QString token=rx.cap(1);
    // See in human format
//    qDebug() << data.replace("&","\n&");

    // Retrieve available videos from &fmt_map argument
    QStringList l;
    int n;
    rx.setPattern("&fmt_list=(.*)&");
    rx.setMinimal(true);
    rx.indexIn(data);
    l=rx.cap(1).split(",");
    for (n=0; n<l.size(); n++)
    {
        mode.fmt=l.at(n).split("/").at(0);
        // http://en.wikipedia.org/wiki/YouTube#Quality_and_codecs
        switch (mode.fmt.toInt()) {
        case 0:
        case 5:
            mode.humanFmt="640x360";
            mode.container="FLV";
            mode.xres=640; mode.yres=360;
            break;
        case 6:
            mode.humanFmt="480x360";
            mode.container="FLV";
            mode.xres=480; mode.yres=360;
            break;
        case 13:
            mode.humanFmt="176x144";
            mode.container="3GP";
            mode.xres=480; mode.yres=360;
            break;
        case 17:
            mode.humanFmt="176x144";
            mode.container="3GP";
            mode.xres=176; mode.yres=144;
            break;
        case 18:
            mode.humanFmt="480x360";
            mode.container="MP4";
            mode.xres=480; mode.yres=360;
            break;
        case 22:
            mode.humanFmt="1280x720";
            mode.container="MP4";
            mode.xres=1280; mode.yres=720;
            break;
        case 34:
            mode.humanFmt="640x480";
            mode.container="FLV";
            mode.xres=640; mode.yres=480;
            break;
        case 35:
            mode.humanFmt="854x480";
            mode.container="FLV";
            mode.xres=854; mode.yres=480;
            break;
        case 37:
            mode.humanFmt="1920x1080";
            mode.container="MP4";
            mode.xres=1920; mode.yres=1080;
            break;
        case 38:
            mode.humanFmt="4096x2304";
            mode.container="MP4";
            mode.xres=4096; mode.yres=2304;
            break;
        case 43:
            mode.humanFmt="640x360";
            mode.container="WebM";
            mode.xres=640; mode.yres=360;
            break;
        case 44:
            mode.humanFmt="854x480";
            mode.container="WebM";
            mode.xres=854; mode.yres=480;
            break;
        case 45:
            mode.humanFmt="1280x720";
            mode.container="WebM";
            mode.xres=1280; mode.yres=720;
            break;
        default:
            mode.humanFmt="unknown mode:"+mode.fmt;
            mode.container="";
            mode.xres=320; mode.yres=240;
            break;
        }
        mode.url_play=getUrlsFromPattern(data, mode.fmt.toInt());
        mode.url_retrieve=mode.url_play;//getUrlsFromPattern(data, mode.fmt.toInt());


        qDebug() << "Video format found: " << mode.fmt << "-" << mode.humanFmt << " with " << mode.container << "format";
        count++;
        tube.video_info.append(mode);
    }
    tube.ID_preferred_video_format=youTube::getPreferredResolution(tube, youTube::ID_preferred_video_format);
    qDebug() << "Using index" << tube.ID_preferred_video_format << " as default";
    return count;
}

// Retrieve url from &fmt_stream_map
// These url are used for download or watching a video in different formats
// url_or_strem=0 search for fmt_strem_map
// url_or_strem=1 search for fmt_url_map
QString getUrlsFromPattern(QByteArray data, int itag)
{
    QString url;
    QRegExp rx;

//    qDebug() << data;
    // Consider data starting above "&url_encoded_fmt_stream_map="
//    rx.setPattern("&itag="+QString::number(itag)+",url=(.*)&type=");
    rx.setPattern("&url_encoded_fmt_stream_map=(.*)");
    rx.setMinimal(false);
    rx.indexIn(data);
    data=rx.cap(1).toAscii();

    rx.setPattern("url=(http:.*&itag=([0-9]{1,}).*&type=.*)&");
    rx.setMinimal(true);
    rx.indexIn(data);

    bool b = false;
    int pos = 0;
    while ((pos = rx.indexIn(data, pos)) != -1 && !b ) {
        if (rx.cap(2).toInt() == itag) {
            b = true;
            url=rx.cap(1);
        }
        pos += rx.matchedLength();
     }

    return url;
}

// Compose a string url with user parameters and
// fill tubes list with the retrieved videos
QList<YouTubeStruct> youTube::getList(QString search,
                                      StandardVideoFeeds feed,
                                      QString videoFeedsCategory,
                                      int start_feed_number, int feeds_to_get,  // start_feed_number > 0
                                      int order,
                                      QString preferred_language,
                                      int time
                                      )
{
    QString getUrl;
    QString feedTime;
    QString orderBy;
    currentSearchText = search;
    switch (time) {
        case 0: feedTime="all_time";   break;
        case 1: feedTime="today";      break;
        case 2: feedTime="this_week";  break;
        case 3: feedTime="this_month"; break;
    }
    switch (order) {
        case 0: orderBy="relevance";   break;
        case 1: orderBy="published";   break;
        case 2: orderBy="viewCount";   break;
        case 3: orderBy="rating";      break;
    }

    switch (feed) {
    case VIDEO_FEED_NONE_USE_SEARCH:
        getUrl="http://gdata.youtube.com/feeds/api/videos?q="+search.replace(" ","+");
        break;
    case VIDEO_FEED_TOP_RATED:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/top_rated?";
        break;
    case VIDEO_FEED_TOP_FAVORITES:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/top_favorites?";
        break;
    case VIDEO_FEED_MOST_VIEWED:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/most_viewed?";
        break;
    case VIDEO_FEED_MOST_POPULAR:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/most_popular?";
        break;
    case VIDEO_FEED_MOST_RECENT:
        getUrl="http://gdata.youtube.com/feeds/api/standardfeeds/most_recent?";
        break;
    }

    getUrl+="&start-index="+QString::number(start_feed_number);
    getUrl+="&max-results="+QString::number(feeds_to_get);
    getUrl+="&orderby="+orderBy;
    if (!preferred_language.isEmpty()) getUrl+="&lr="+preferred_language.left(2);
    if (feed==VIDEO_FEED_NONE_USE_SEARCH) {
        if (!feedTime.isEmpty()) getUrl+="&time="+feedTime;
        if (!videoFeedsCategory.isEmpty()) getUrl+="&category="+videoFeedsCategory;
    }
    getUrl+="&v=1";

    return fillSearchTubes(getUrl);
}

// Make a specific search
QList<YouTubeStruct> youTube::fillSearchTubes(QString getUrl)
{
    QByteArray data;
    YouTubeStruct tube;
    QList<YouTubeStruct> tubes;

    network *net = new network(false);
    data=net->get_synchronous(getUrl,false);

//    qDebug() << "\n\n\nGetting feeds:" << getUrl;
//    qDebug() << data;
    net->deleteLater();

    if (data.isEmpty()) return tubes;

    QDomDocument doc("mydocument");
    doc.setContent(data);
    QDomElement docElem = doc.documentElement();

    QDomNode n = docElem.firstChild();
    QDomElement e,t;

    // Parse all <entry> tag to get infos about the videos
    tubes.clear();
    while(!n.isNull()) {
        e = n.toElement(); // try to convert the node to an element.
        if(!e.isNull()) {
            if (e.tagName()=="entry")
            {
                tube.id          = e.firstChildElement("id").text();
                tube.published   = e.firstChildElement("published").text().replace("T"," ").replace("Z","");
                tube.updated     = e.firstChildElement("updated").text().replace("T"," ").replace("Z","");
                tube.title       = e.firstChildElement("title").text();
                tube.content     = e.firstChildElement("content").text();
                tube.author_name = e.firstChildElement("author").firstChildElement("name").text();
                tube.author_uri  = e.firstChildElement("author").firstChildElement("uri").text();

                tube.duration    = e.firstChildElement("media:group").firstChildElement("yt:duration").attribute("seconds").toAscii();
                tube.mediaUrl    = e.firstChildElement("media:group").firstChildElement("media:player").attribute("url").toAscii().replace("&amp","&");
                t = e.firstChildElement("media:group").firstChildElement("media:thumbnail");
                tube.p[0].height = t.attribute("height").toInt();
                tube.p[0].width  = t.attribute("width").toInt();
                tube.p[0].url    = t.attribute("url").toAscii();
                t=t.nextSiblingElement("media:thumbnail");
                tube.p[1].height = t.attribute("height").toInt();
                tube.p[1].width  = t.attribute("width").toInt();
                tube.p[1].url    = t.attribute("url").toAscii();
                t=t.nextSiblingElement("media:thumbnail");
                tube.p[2].height = t.attribute("height").toInt();
                tube.p[2].width  = t.attribute("width").toInt();
                tube.p[2].url    = t.attribute("url").toAscii();
                t=t.nextSiblingElement("media:thumbnail");
                tube.p[3].height = t.attribute("height").toInt();
                tube.p[3].width  = t.attribute("width").toInt();
                tube.p[3].url    = t.attribute("url").toAscii();

                tube.rating_average       = e.firstChildElement("gd:rating").attribute("average").toAscii();
                tube.rating_numRaters     = e.firstChildElement("gd:rating").attribute("numRaters").toAscii();
                tube.statistics_viewCount = e.firstChildElement("yt:statistics").attribute("viewCount").toAscii();
                tube.video_id=tube.id.right(11);
                tube.gotWithSearchText    = currentSearchText;
                tubes.append(tube);
            }
        }
        n = n.nextSibling();
    }
    return tubes;
}

